# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any

from pyhocon import ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.rest_api.query_merger import QueryMerger

LOGGER = logging.getLogger(__name__)


class ModeDashboardUsageExtractor(Extractor):
    """
    A Extractor that extracts Mode dashboard's accumulated view count
    """

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(restapi_query=restapi_query,
                                                                            conf=self._conf)

    def extract(self) -> Any:
        return self._extractor.extract()

    def get_scope(self) -> str:
        return 'extractor.mode_dashboard_usage'

    def _build_restapi_query(self) -> ModePaginatedRestApiQuery:
        """
        Build REST API Query. To get Mode Dashboard usage, it needs to call three discovery APIs (
        spaces API, reports API and report stats API).
        :return: A RestApiQuery that provides Mode Dashboard metadata
        """

        seed_query = ModeDashboardUtils.get_seed_query(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf, discover_auth=True)

        # Reports
        # https://mode.com/developer/discovery-api/analytics/reports/
        reports_url = 'https://app.mode.com/batch/{organization}/reports'
        reports_json_path = 'reports[*].[token, space_token]'
        reports_field_names = ['dashboard_id', 'dashboard_group_id']
        reports_max_record_size = 1000
        reports_pagination_json_path = 'reports[*]'
        spaces_query = ModeDashboardUtils.get_spaces_query_api(conf=self._conf)
        spaces_query_merger = QueryMerger(query_to_merge=spaces_query, merge_key='dashboard_group_id')
        reports_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=reports_url, params=params,
                                                  json_path=reports_json_path, field_names=reports_field_names,
                                                  skip_no_result=True, max_record_size=reports_max_record_size,
                                                  pagination_json_path=reports_pagination_json_path,
                                                  query_merger=spaces_query_merger)

        # https://mode.com/developer/discovery-api/analytics/report-stats/
        stats_url = 'https://app.mode.com/batch/{organization}/report_stats'
        stats_json_path = 'report_stats[*].[report_token, view_count]'
        stats_field_names = ['dashboard_id', 'accumulated_view_count']
        stats_max_record_size = 1000
        stats_pagination_json_path = 'report_stats[*]'
        reports_query_merger = QueryMerger(query_to_merge=reports_query, merge_key='dashboard_id')
        report_stats_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=stats_url, params=params,
                                                       json_path=stats_json_path, field_names=stats_field_names,
                                                       skip_no_result=True, max_record_size=stats_max_record_size,
                                                       pagination_json_path=stats_pagination_json_path,
                                                       query_merger=reports_query_merger)

        return report_stats_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Union

from pyparsing import (
    Forward, Group, Keyword, OneOrMore, Optional, Word, alphanums, delimitedList, nestedExpr, nums, originalTextFor,
)

from databuilder.models.table_metadata import ColumnMetadata
from databuilder.models.type_metadata import (
    ArrayTypeMetadata, MapTypeMetadata, ScalarTypeMetadata, StructTypeMetadata, TypeMetadata,
)

array_keyword = Keyword("array")
map_keyword = Keyword("map")
struct_keyword = Keyword("row")

field_name = Word(alphanums + "_")
field_type = Forward()

# Scalar types
scalar_quantifier = "(" + Word(nums) + Optional(")" | "," + Word(nums) + ")")
scalar_type = OneOrMore(Word(alphanums + "_")) + Optional(scalar_quantifier)

# Complex types
array_field = "(" + field_type("type")
map_field = originalTextFor(scalar_type)("key") + "," + field_type("type")
struct_field = field_name("name") + field_type("type")
struct_list = delimitedList(Group(struct_field))
array_type = nestedExpr(
    opener=array_keyword, closer=")", content=array_field, ignoreExpr=None
)
map_type = nestedExpr(
    opener=map_keyword + "(", closer=")", content=map_field, ignoreExpr=None
)
struct_type = nestedExpr(
    opener=struct_keyword + "(", closer=")", content=struct_list, ignoreExpr=None
)

field_type <<= originalTextFor(array_type | map_type | struct_type | scalar_type)

complex_type = (array_type("array_type") | map_type("map_type") | struct_type("struct_type") |
                scalar_type("scalar_type"))


def parse_trino_type(type_str: str, name: str, parent: Union[ColumnMetadata, TypeMetadata]) -> TypeMetadata:
    type_str = type_str.lower()
    type_str = type_str.replace('\"', '')  # Remove quotes around names that are added when querying from HMS
    parsed_type = complex_type.parseString(type_str, parseAll=True)

    if parsed_type.scalar_type:
        return ScalarTypeMetadata(name=name,
                                  parent=parent,
                                  type_str=type_str)

    results = parsed_type[0]
    if parsed_type.array_type:
        array_type_metadata = ArrayTypeMetadata(name=name,
                                                parent=parent,
                                                type_str=type_str)
        array_inner_type = parse_trino_type(results.type, '_inner_', array_type_metadata)
        if not isinstance(array_inner_type, ScalarTypeMetadata):
            array_type_metadata.array_inner_type = array_inner_type
        return array_type_metadata
    elif parsed_type.map_type:
        map_type_metadata = MapTypeMetadata(name=name,
                                            parent=parent,
                                            type_str=type_str)
        map_type_metadata.map_key_type = parse_trino_type(results.key, '_map_key', map_type_metadata)
        map_type_metadata.map_value_type = parse_trino_type(results.type, '_map_value', map_type_metadata)
        return map_type_metadata
    elif parsed_type.struct_type:
        struct_type_metadata = StructTypeMetadata(name=name,
                                                  parent=parent,
                                                  type_str=type_str)
        struct_items = {}
        for index, result in enumerate(results):
            struct_items[result.name] = parse_trino_type(result.type, result.name, struct_type_metadata)
            struct_items[result.name].sort_order = index

        struct_type_metadata.struct_items = struct_items
        return struct_type_metadata
    else:
        raise Exception(f"Unrecognized type: {type_str}")



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Dict, Iterator, List, Optional, Set, Tuple, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasDashboardTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import (
    Dashboard as RDSDashboard, DashboardCluster as RDSDashboardCluster, DashboardDescription as RDSDashboardDescription,
    DashboardGroup as RDSDashboardGroup, DashboardGroupDescription as RDSDashboardGroupDescription,
    DashboardTag as RDSDashboardTag,
)
from amundsen_rds.models.tag import Tag as RDSTag

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.cluster import cluster_constants
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
# TODO: We could separate TagMetadata from table_metadata to own module
from databuilder.models.table_metadata import TagMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import (
    add_entity_relationship, get_entity_attrs, get_entity_relationships,
)
from databuilder.utils.atlas import AtlasSerializedEntityOperation


class DashboardMetadata(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    Dashboard metadata including dashboard group name, dashboardgroup description, dashboard description,
    and tags.

    Some other metadata e.g. Owners and last-reload/modified times are provided by other models
    e.g. DashboardOwner

    It implements Neo4jCsvSerializable so that it can be serialized to produce
    Dashboard, Tag, Description and relations between those. Additionally, it will create a
    Dashboardgroup with relationships to the Dashboard.
    """
    CLUSTER_KEY_FORMAT = '{product}_dashboard://{cluster}'
    CLUSTER_DASHBOARD_GROUP_RELATION_TYPE = 'DASHBOARD_GROUP'
    DASHBOARD_GROUP_CLUSTER_RELATION_TYPE = 'DASHBOARD_GROUP_OF'

    DASHBOARD_NODE_LABEL = 'Dashboard'
    DASHBOARD_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group}/{dashboard_name}'
    DASHBOARD_NAME = 'name'
    DASHBOARD_CREATED_TIME_STAMP = 'created_timestamp'
    DASHBOARD_GROUP_URL = 'dashboard_group_url'
    DASHBOARD_URL = 'dashboard_url'

    DASHBOARD_DESCRIPTION_NODE_LABEL = 'Description'
    DASHBOARD_DESCRIPTION = 'description'
    DASHBOARD_DESCRIPTION_FORMAT = \
        '{product}_dashboard://{cluster}.{dashboard_group}/{dashboard_name}/_description'
    DASHBOARD_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_DASHBOARD_RELATION_TYPE = 'DESCRIPTION_OF'

    DASHBOARD_GROUP_NODE_LABEL = 'Dashboardgroup'
    DASHBOARD_GROUP_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group}'
    DASHBOARD_GROUP_DASHBOARD_RELATION_TYPE = 'DASHBOARD'
    DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE = 'DASHBOARD_OF'

    DASHBOARD_GROUP_DESCRIPTION_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group}/_description'

    DASHBOARD_TAG_RELATION_TYPE = 'TAG'
    TAG_DASHBOARD_RELATION_TYPE = 'TAG_OF'

    ATLAS_DASHBOARD_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group_id}/{dashboard_id}'
    ATLAS_DASHBOARD_GROUP_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group_id}'

    serialized_nodes: Set[Any] = set()
    serialized_rels: Set[Any] = set()

    def __init__(self,
                 dashboard_group: str,
                 dashboard_name: str,
                 description: Union[str, None],
                 tags: List = None,
                 cluster: str = 'gold',
                 product: Optional[str] = '',
                 dashboard_group_id: Optional[str] = None,
                 dashboard_id: Optional[str] = None,
                 dashboard_group_description: Optional[str] = None,
                 created_timestamp: Optional[int] = None,
                 dashboard_group_url: Optional[str] = None,
                 dashboard_url: Optional[str] = None,
                 **kwargs: Any
                 ) -> None:

        self.dashboard_group = dashboard_group
        self.dashboard_name = dashboard_name
        self.dashboard_group_id = dashboard_group_id if dashboard_group_id else dashboard_group
        self.dashboard_id = dashboard_id if dashboard_id else dashboard_name
        self.description = description
        self.tags = tags
        self.product = product
        self.cluster = cluster
        self.dashboard_group_description = dashboard_group_description
        self.created_timestamp = created_timestamp
        self.dashboard_group_url = dashboard_group_url
        self.dashboard_url = dashboard_url
        self._processed_cluster: Set[str] = set()
        self._processed_dashboard_group: Set[str] = set()
        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()
        self._record_iterator = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def __repr__(self) -> str:
        return f'DashboardMetadata(' \
               f'{self.dashboard_group!r}, {self.dashboard_name!r}, {self.description!r}, {self.tags!r}, ' \
               f'{self.dashboard_group_id!r}, {self.dashboard_id!r}, {self.dashboard_group_description!r}, ' \
               f'{self.created_timestamp!r}, {self.dashboard_group_url!r}, {self.dashboard_url!r})'

    def _get_cluster_key(self) -> str:
        return DashboardMetadata.CLUSTER_KEY_FORMAT.format(cluster=self.cluster,
                                                           product=self.product)

    def _get_dashboard_key(self) -> str:
        return DashboardMetadata.DASHBOARD_KEY_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                             dashboard_name=self.dashboard_id,
                                                             cluster=self.cluster,
                                                             product=self.product)

    def _get_dashboard_description_key(self) -> str:
        return DashboardMetadata.DASHBOARD_DESCRIPTION_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                                     dashboard_name=self.dashboard_id,
                                                                     cluster=self.cluster,
                                                                     product=self.product)

    def _get_dashboard_group_description_key(self) -> str:
        return DashboardMetadata.DASHBOARD_GROUP_DESCRIPTION_KEY_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                                               cluster=self.cluster,
                                                                               product=self.product)

    def _get_dashboard_group_key(self) -> str:
        return DashboardMetadata.DASHBOARD_GROUP_KEY_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                                   cluster=self.cluster,
                                                                   product=self.product)

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:

        # dashboard group
        group_attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_dashboard_group_key()),
            ('name', self.dashboard_group),
            ('id', self.dashboard_group_id),
            ('description', self.dashboard_group_description),
            ('url', self.dashboard_group_url),
        ]
        dashboard_group_entity_attrs = get_entity_attrs(group_attrs_mapping)

        dashboard_group_entity = AtlasEntity(
            typeName=AtlasDashboardTypes.group,
            operation=AtlasSerializedEntityOperation.CREATE,
            relationships=None,
            attributes=dashboard_group_entity_attrs,
        )

        yield dashboard_group_entity

        # dashboard
        attrs_mapping: List[Tuple[Any, Any]] = [
            (AtlasCommonParams.qualified_name, self._get_dashboard_key()),
            ('name', self.dashboard_name),
            ('description', self.description),
            ('url', self.dashboard_url),
            ('cluster', self.cluster),
            ('product', self.product),
            (AtlasCommonParams.created_timestamp, self.created_timestamp),
        ]

        dashboard_entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'group',
            AtlasDashboardTypes.group,
            self._get_dashboard_group_key(),
        )

        dashboard_entity = AtlasEntity(
            typeName=AtlasDashboardTypes.metadata,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=dashboard_entity_attrs,
            relationships=get_entity_relationships(relationship_list),
        )
        yield dashboard_entity

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_next_node(self) -> Iterator[GraphNode]:
        # Cluster node
        if not self._get_cluster_key() in self._processed_cluster:
            self._processed_cluster.add(self._get_cluster_key())
            cluster_node = GraphNode(
                key=self._get_cluster_key(),
                label=cluster_constants.CLUSTER_NODE_LABEL,
                attributes={
                    cluster_constants.CLUSTER_NAME_PROP_KEY: self.cluster
                }
            )
            yield cluster_node

        # Dashboard node attributes
        dashboard_node_attributes: Dict[str, Any] = {
            DashboardMetadata.DASHBOARD_NAME: self.dashboard_name,
        }
        if self.created_timestamp:
            dashboard_node_attributes[DashboardMetadata.DASHBOARD_CREATED_TIME_STAMP] = self.created_timestamp

        if self.dashboard_url:
            dashboard_node_attributes[DashboardMetadata.DASHBOARD_URL] = self.dashboard_url

        dashboard_node = GraphNode(
            key=self._get_dashboard_key(),
            label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            attributes=dashboard_node_attributes
        )

        yield dashboard_node

        # Dashboard group
        if self.dashboard_group and not self._get_dashboard_group_key() in self._processed_dashboard_group:
            self._processed_dashboard_group.add(self._get_dashboard_group_key())
            dashboard_group_node_attributes = {
                DashboardMetadata.DASHBOARD_NAME: self.dashboard_group,
            }

            if self.dashboard_group_url:
                dashboard_group_node_attributes[DashboardMetadata.DASHBOARD_GROUP_URL] = self.dashboard_group_url

            dashboard_group_node = GraphNode(
                key=self._get_dashboard_group_key(),
                label=DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
                attributes=dashboard_group_node_attributes
            )

            yield dashboard_group_node

        # Dashboard group description
        if self.dashboard_group_description:
            dashboard_group_description_node = GraphNode(
                key=self._get_dashboard_group_description_key(),
                label=DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                attributes={
                    DashboardMetadata.DASHBOARD_DESCRIPTION: self.dashboard_group_description
                }
            )
            yield dashboard_group_description_node

        # Dashboard description node
        if self.description:
            dashboard_description_node = GraphNode(
                key=self._get_dashboard_description_key(),
                label=DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                attributes={
                    DashboardMetadata.DASHBOARD_DESCRIPTION: self.description
                }
            )
            yield dashboard_description_node

        # Dashboard tag node
        if self.tags:
            for tag in self.tags:
                dashboard_tag_node = GraphNode(
                    key=TagMetadata.get_tag_key(tag),
                    label=TagMetadata.TAG_NODE_LABEL,
                    attributes={
                        TagMetadata.TAG_TYPE: 'dashboard'
                    }
                )
                yield dashboard_tag_node

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self) -> Iterator[GraphRelationship]:
        # Cluster <-> Dashboard group
        cluster_dashboard_group_relationship = GraphRelationship(
            start_label=cluster_constants.CLUSTER_NODE_LABEL,
            start_key=self._get_cluster_key(),
            end_label=DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
            end_key=self._get_dashboard_group_key(),
            type=DashboardMetadata.CLUSTER_DASHBOARD_GROUP_RELATION_TYPE,
            reverse_type=DashboardMetadata.DASHBOARD_GROUP_CLUSTER_RELATION_TYPE,
            attributes={}
        )
        yield cluster_dashboard_group_relationship

        # Dashboard group > Dashboard group description relation
        if self.dashboard_group_description:
            dashboard_group_description_relationship = GraphRelationship(
                start_label=DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
                start_key=self._get_dashboard_group_key(),
                end_label=DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                end_key=self._get_dashboard_group_description_key(),
                type=DashboardMetadata.DASHBOARD_DESCRIPTION_RELATION_TYPE,
                reverse_type=DashboardMetadata.DESCRIPTION_DASHBOARD_RELATION_TYPE,
                attributes={}
            )
            yield dashboard_group_description_relationship

        # Dashboard group > Dashboard relation
        dashboard_group_dashboard_relationship = GraphRelationship(
            start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            end_label=DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
            start_key=self._get_dashboard_key(),
            end_key=self._get_dashboard_group_key(),
            type=DashboardMetadata.DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE,
            reverse_type=DashboardMetadata.DASHBOARD_GROUP_DASHBOARD_RELATION_TYPE,
            attributes={}
        )
        yield dashboard_group_dashboard_relationship

        # Dashboard > Dashboard description relation
        if self.description:
            dashboard_description_relationship = GraphRelationship(
                start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
                end_label=DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                start_key=self._get_dashboard_key(),
                end_key=self._get_dashboard_description_key(),
                type=DashboardMetadata.DASHBOARD_DESCRIPTION_RELATION_TYPE,
                reverse_type=DashboardMetadata.DESCRIPTION_DASHBOARD_RELATION_TYPE,
                attributes={}
            )
            yield dashboard_description_relationship

        # Dashboard > Dashboard tag relation
        if self.tags:
            for tag in self.tags:
                dashboard_tag_relationship = GraphRelationship(
                    start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
                    end_label=TagMetadata.TAG_NODE_LABEL,
                    start_key=self._get_dashboard_key(),
                    end_key=TagMetadata.get_tag_key(tag),
                    type=DashboardMetadata.DASHBOARD_TAG_RELATION_TYPE,
                    reverse_type=DashboardMetadata.TAG_DASHBOARD_RELATION_TYPE,
                    attributes={}
                )
                yield dashboard_tag_relationship

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        # Cluster
        if not self._get_cluster_key() in self._processed_cluster:
            self._processed_cluster.add(self._get_cluster_key())
            yield RDSDashboardCluster(
                rk=self._get_cluster_key(),
                name=self.cluster
            )

        # Dashboard group
        if self.dashboard_group and not self._get_dashboard_group_key() in self._processed_dashboard_group:
            self._processed_dashboard_group.add(self._get_dashboard_group_key())
            dashboard_group_record = RDSDashboardGroup(
                rk=self._get_dashboard_group_key(),
                name=self.dashboard_group,
                cluster_rk=self._get_cluster_key()
            )
            if self.dashboard_group_url:
                dashboard_group_record.dashboard_group_url = self.dashboard_group_url

            yield dashboard_group_record

        # Dashboard group description
        if self.dashboard_group_description:
            yield RDSDashboardGroupDescription(
                rk=self._get_dashboard_group_description_key(),
                description=self.dashboard_group_description,
                dashboard_group_rk=self._get_dashboard_group_key()
            )

        # Dashboard
        dashboard_record = RDSDashboard(
            rk=self._get_dashboard_key(),
            name=self.dashboard_name,
            dashboard_group_rk=self._get_dashboard_group_key()
        )
        if self.created_timestamp:
            dashboard_record.created_timestamp = self.created_timestamp

        if self.dashboard_url:
            dashboard_record.dashboard_url = self.dashboard_url

        yield dashboard_record

        # Dashboard description
        if self.description:
            yield RDSDashboardDescription(
                rk=self._get_dashboard_description_key(),
                description=self.description,
                dashboard_rk=self._get_dashboard_key()
            )

        # Dashboard tag
        if self.tags:
            for tag in self.tags:
                tag_record = RDSTag(
                    rk=TagMetadata.get_tag_key(tag),
                    tag_type='dashboard',
                )
                yield tag_record

                dashboard_tag_record = RDSDashboardTag(
                    dashboard_rk=self._get_dashboard_key(),
                    tag_rk=TagMetadata.get_tag_key(tag)
                )
                yield dashboard_tag_record



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import Iterator, Union

from feast import FeatureStore, FeatureView
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class FeastExtractor(Extractor):
    """
    Extracts feature tables from Feast feature store file. Since Feast is
    a metadata store (and not the database itself), it maps the
    following attributes:

     * a database is name of feast project
     * table name is a name of the feature view
     * columns are features stored in the feature view
    """

    FEAST_REPOSITORY_PATH = "/path/to/repository"
    DESCRIBE_FEATURE_VIEWS = "describe_feature_views"
    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {FEAST_REPOSITORY_PATH: ".", DESCRIBE_FEATURE_VIEWS: True}
    )

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(FeastExtractor.DEFAULT_CONFIG)
        self._feast_repository_path = conf.get_string(
            FeastExtractor.FEAST_REPOSITORY_PATH
        )
        self._describe_feature_views = conf.get_bool(
            FeastExtractor.DESCRIBE_FEATURE_VIEWS
        )
        self._feast = FeatureStore(repo_path=self._feast_repository_path)
        self._extract_iter: Union[None, Iterator] = None

    def get_scope(self) -> str:
        return "extractor.feast"

    def extract(self) -> Union[TableMetadata, None]:
        """
        For every feature table from Feast, a multiple objets are extracted:

        1. TableMetadata with feature view description
        2. Programmatic Description of the feature view, containing
           metadata - date of creation and labels
        3. Programmatic Description with Batch Source specification
        4. (if applicable) Programmatic Description with Stream Source
           specification
        """
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _get_extract_iter(self) -> Iterator[TableMetadata]:
        for feature_view in self._feast.list_feature_views():
            yield from self._extract_feature_view(feature_view)

    def _extract_feature_view(
        self, feature_view: FeatureView
    ) -> Iterator[TableMetadata]:
        columns = []
        for index, entity_name in enumerate(feature_view.entities):
            entity = self._feast.get_entity(entity_name)
            columns.append(
                ColumnMetadata(
                    entity.name, entity.description, entity.value_type.name, index
                )
            )

        for index, feature in enumerate(feature_view.features):
            columns.append(
                ColumnMetadata(
                    feature.name,
                    None,
                    feature.dtype.name,
                    len(feature_view.entities) + index,
                )
            )

        yield TableMetadata(
            "feast",
            self._feast.config.provider,
            self._feast.project,
            feature_view.name,
            None,
            columns,
        )

        if self._describe_feature_views:
            description = str()
            if feature_view.created_timestamp:
                created_at = datetime.utcfromtimestamp(
                    feature_view.created_timestamp.timestamp()
                )
                description = f"* Created at **{created_at}**\n"

            if feature_view.tags:
                description += "* Tags:\n"
                for key, value in feature_view.tags.items():
                    description += f"    * {key}: **{value}**\n"

            yield TableMetadata(
                "feast",
                self._feast.config.provider,
                self._feast.project,
                feature_view.name,
                description,
                description_source="feature_view_details",
            )

            yield TableMetadata(
                "feast",
                self._feast.config.provider,
                self._feast.project,
                feature_view.name,
                f"```\n{str(feature_view.batch_source.to_proto())}```",
                description_source="batch_source",
            )

            if feature_view.stream_source:
                yield TableMetadata(
                    "feast",
                    self._feast.config.provider,
                    self._feast.project,
                    feature_view.name,
                    f"```\n{str(feature_view.stream_source.to_proto())}```",
                    description_source="stream_source",
                )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from datetime import date
from typing import (
    Any, Generator, List,
)
from uuid import uuid4

from elasticsearch.exceptions import NotFoundError
from elasticsearch.helpers import parallel_bulk
from elasticsearch_dsl.connections import Connections, connections
from elasticsearch_dsl.document import Document
from elasticsearch_dsl.index import Index
from pyhocon import ConfigTree

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.task.base_task import Task
from databuilder.task.search.document_mappings import RESOURCE_TO_MAPPING, SearchableResource
from databuilder.transformer.base_transformer import NoopTransformer, Transformer
from databuilder.utils.closer import Closer

LOGGER = logging.getLogger(__name__)


class SearchMetadatatoElasticasearchTask(Task):

    ENTITY_TYPE = 'doc_type'
    ELASTICSEARCH_CLIENT_CONFIG_KEY = 'client'
    MAPPING_CLASS = 'document_mapping'
    ELASTICSEARCH_ALIAS_CONFIG_KEY = 'alias'
    ELASTICSEARCH_NEW_INDEX = 'new_index'
    ELASTICSEARCH_PUBLISHER_BATCH_SIZE = 'batch_size'
    ELASTICSEARCH_TIMEOUT_SEC = 'es_timeout_sec'
    DATE = 'date'

    today = date.today().strftime("%Y%m%d")

    def __init__(self,
                 extractor: Extractor,
                 transformer: Transformer = NoopTransformer()) -> None:
        self.extractor = extractor
        self.transformer = transformer

        self._closer = Closer()
        self._closer.register(self.extractor.close)
        self._closer.register(self.transformer.close)

    def init(self, conf: ConfigTree) -> None:
        # initialize extractor with configurarion
        self.extractor.init(Scoped.get_scoped_conf(conf, self.extractor.get_scope()))
        # initialize transformer with configuration
        self.transformer.init(Scoped.get_scoped_conf(conf, self.transformer.get_scope()))

        # task configuration
        conf = Scoped.get_scoped_conf(conf, self.get_scope())
        self.date = conf.get_string(SearchMetadatatoElasticasearchTask.DATE, self.today)
        self.entity = conf.get_string(SearchMetadatatoElasticasearchTask.ENTITY_TYPE).lower()
        self.elasticsearch_client = conf.get(
            SearchMetadatatoElasticasearchTask.ELASTICSEARCH_CLIENT_CONFIG_KEY
        )
        self.elasticsearch_alias = conf.get(
            SearchMetadatatoElasticasearchTask.ELASTICSEARCH_ALIAS_CONFIG_KEY
        )
        self.elasticsearch_new_index = conf.get(
            SearchMetadatatoElasticasearchTask.ELASTICSEARCH_NEW_INDEX,
            self.create_new_index_name())
        self.document_mapping = conf.get(SearchMetadatatoElasticasearchTask.MAPPING_CLASS,
                                         RESOURCE_TO_MAPPING[self.entity])

        if not issubclass(self.document_mapping, SearchableResource):
            msg = "Provided document_mapping should be instance" \
                f" of SearchableResource not {type(self.document_mapping)}"
            LOGGER.error(msg)
            raise TypeError(msg)

        self.elasticsearch_batch_size = conf.get(
            SearchMetadatatoElasticasearchTask.ELASTICSEARCH_PUBLISHER_BATCH_SIZE, 10000
        )
        self.elasticsearch_timeout_sec = conf.get(
            SearchMetadatatoElasticasearchTask.ELASTICSEARCH_TIMEOUT_SEC, 120
        )

    def create_new_index_name(self) -> str:
        hex_string = uuid4().hex
        return f"{self.elasticsearch_alias}_{self.date}_{hex_string}"

    def to_document(self, metadata: Any) -> Document:
        return self.document_mapping(_index=self.elasticsearch_new_index,
                                     **metadata)

    def generate_documents(self, record: Any) -> Generator:
        # iterate through records
        while record:
            record = self.transformer.transform(record)
            if not record:
                # Move on if the transformer filtered the record out
                record = self.extractor.extract()
                continue
            document = self.to_document(metadata=record).to_dict(True)
            document['_source']['resource_type'] = self.entity

            yield document
            record = self.extractor.extract()

    def _get_old_index(self, connection: Connections) -> List[str]:
        """
        Retrieve all indices that currently have {elasticsearch_alias} alias
        :return: list of elasticsearch indices
        """
        try:
            indices = connection.indices.get_alias(self.elasticsearch_alias).keys()
            return indices
        except NotFoundError:
            LOGGER.warn("Received index not found error from Elasticsearch. " +
                        "The index doesn't exist for a newly created ES. It's OK on first run.")
            # return empty list on exception
            return []

    def _delete_old_index(self, connection: Connections, document_index: Index) -> None:
        alias_updates = []
        previous_index = self._get_old_index(connection=connection)
        for previous_index_name in previous_index:
            if previous_index_name != document_index._name:
                LOGGER.info(f"Deleting old index {previous_index_name}")
                alias_updates.append({"remove_index": {"index": previous_index_name}})
        alias_updates.append({"add": {
            "index": self.elasticsearch_new_index,
            "alias": self.elasticsearch_alias}})
        connection.indices.update_aliases({"actions": alias_updates})

    def run(self) -> None:
        LOGGER.info('Running search metadata to Elasticsearch task')
        try:
            # extract records from metadata store
            record = self.extractor.extract()

            # create connection
            connections.add_connection('default', self.elasticsearch_client)
            connection = connections.get_connection()

            # health check ES
            health = connection.cluster.health()
            status = health["status"]
            if status not in ("green", "yellow"):
                msg = f"Elasticsearch healthcheck failed: {status}"
                LOGGER.error(msg)
                raise Exception(msg)

            # create index
            LOGGER.info(f"Creating ES index {self.elasticsearch_new_index}")
            index = Index(name=self.elasticsearch_new_index, using=self.elasticsearch_client)
            index.document(self.document_mapping)

            # allow for longer ngram length
            index.settings(max_shingle_diff=10)

            index.create()

            # publish search metadata to ES
            cnt = 0
            for success, info in parallel_bulk(connection,
                                               self.generate_documents(record=record),
                                               raise_on_error=False,
                                               chunk_size=self.elasticsearch_batch_size,
                                               request_timeout=self.elasticsearch_timeout_sec):
                if not success:
                    LOGGER.warn(f"There was an error while indexing a document to ES: {info}")
                else:
                    cnt += 1
                if cnt == self.elasticsearch_batch_size:
                    LOGGER.info(f'Published {str(cnt*self.elasticsearch_batch_size)} records to ES')

            # delete old index
            self._delete_old_index(connection=connection,
                                   document_index=index)

            LOGGER.info("Elasticsearch Indexing completed")
        finally:
            self._closer.close()

    def get_scope(self) -> str:
        return 'task.search_metadata_to_elasticsearch'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

CLUSTER_NODE_LABEL = 'Cluster'

CLUSTER_RELATION_TYPE = 'CLUSTER'
CLUSTER_REVERSE_RELATION_TYPE = 'CLUSTER_OF'

CLUSTER_NAME_PROP_KEY = 'name'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Dict

from elasticsearch_dsl import (
    Date, Document, Keyword, MetaField, RankFeatures, Text, analysis, token_filter, tokenizer,
)

POSITIONS_OFFSETS = "with_positions_offsets"


class Tokenizer:
    # separate tokens on all non-alphanumeric characters and whitespace
    alphanum_tokenizer = tokenizer("alphanum_tokenizer",
                                   type="pattern",
                                   pattern="[^a-zA-Z0-9]")


class Filter:
    english_stop = token_filter("english_stop", type="stop", stopwords="_english_")

    # uses default porter stemmer
    english_stemmer = token_filter("english_stemmer", type="stemmer", language="english")

    english_possessive_stemmer = token_filter("english_possessive_stemmer",
                                              type="stemmer",
                                              language="possessive_english")


class Analyzer:
    # tokenizes and makes the tokens lowercase
    general_analyzer = analysis.analyzer("general_analyzer",
                                         tokenizer=Tokenizer.alphanum_tokenizer,
                                         filter=["lowercase"])

    # provides light stemming for english tokens
    stemming_analyzer = analysis.analyzer("stemming_analyzer",
                                          tokenizer=Tokenizer.alphanum_tokenizer,
                                          filter=["lowercase", "kstem"])

    # uses grammar based tokenization before analysis (e.g. "it's fine" -> ["it's", "fine"])
    english_analyzer = analysis.analyzer("english_analyzer",
                                         tokenizer=tokenizer("standard_tokenizer",
                                                             type="standard"),
                                         filter=[Filter.english_possessive_stemmer,
                                                 "lowercase",
                                                 Filter.english_stop,
                                                 Filter.english_stemmer])

    # tokenizes for words and numbers, removing all other characters before analysis
    # (e.g. "it's fine" -> ["it", "s", "fine"] or "hello_word" -> ["hello", "world"])
    alphanum_analyzer = analysis.analyzer("alphanum_analyzer",
                                          tokenizer=Tokenizer.alphanum_tokenizer,
                                          filter=[Filter.english_possessive_stemmer,
                                                  "lowercase",
                                                  Filter.english_stop,
                                                  Filter.english_stemmer])


class Subfield:
    # combinations of field types and analyzers for additional index time analysis

    keyword = Keyword()

    alphanumeric = Text(analyzer=Analyzer.alphanum_analyzer,
                        term_vector=POSITIONS_OFFSETS)

    alphanumeric_multi = Text(multi=True,
                              analyzer=Analyzer.alphanum_analyzer,
                              term_vector=POSITIONS_OFFSETS)

    general = Text(analyzer=Analyzer.general_analyzer,
                   term_vector=POSITIONS_OFFSETS)

    general_multi = Text(multi=True,
                         analyzer=Analyzer.general_analyzer,
                         term_vector=POSITIONS_OFFSETS)

    @staticmethod
    def get_ngram_subfield(field_name: str,
                           multi: bool = False,
                           min_shingle_size: int = 2,
                           max_shingle_size: int = 5,
                           token_separator: str = ' ') -> Text:

        # using shingle token filter for word level ngrams
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-shingle-tokenfilter.html
        shingle_filter = token_filter(f"shingle_filter_{field_name}",
                                      type="shingle",
                                      output_unigrams=True,
                                      min_shingle_size=min_shingle_size,
                                      max_shingle_size=max_shingle_size,
                                      token_separator=token_separator)

        ngram_analyzer = analysis.analyzer(f"ngram_analyzer_{field_name}",
                                           tokenizer=Tokenizer.alphanum_tokenizer,
                                           filter=[shingle_filter])

        return Text(multi=multi,
                    analyzer=ngram_analyzer,
                    term_vector=POSITIONS_OFFSETS)


# Resource Mappings

# Note: the current analyzers don't support tokenizing on camelcase text, if this is a requirement extend
# these classes and write a custom analyzer
# https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-pattern-analyzer.html


class SearchableResource(Document):
    # For better understanding of field type rationale read "Mapping unstructured content"
    # https://www.elastic.co/guide/en/elasticsearch/reference/current/keyword.html#wildcard-field-type
    key = Text(required=True,
               fields={"keyword": Subfield.keyword},
               analyzer=Analyzer.general_analyzer,
               term_vector=POSITIONS_OFFSETS)
    name = Text(required=True,
                fields={
                    "keyword": Subfield.keyword,
                    "general": Subfield.general,
                    "ngram": Subfield.get_ngram_subfield(
                        field_name="resource_name",
                        max_shingle_size=8
                    )
                },
                analyzer=Analyzer.stemming_analyzer,
                term_vector=POSITIONS_OFFSETS)
    description = Text(analyzer=Analyzer.english_analyzer,
                       fields={
                           "alphanumeric": Subfield.alphanumeric,
                           "general": Subfield.general
                       },
                       term_vector=POSITIONS_OFFSETS)
    badges = Text(multi=True,
                  fields={"keyword": Subfield.keyword},
                  analyzer=Analyzer.general_analyzer,
                  term_vector=POSITIONS_OFFSETS)
    tags = Text(multi=True,
                fields={"keyword": Subfield.keyword},
                analyzer=Analyzer.general_analyzer,
                term_vector=POSITIONS_OFFSETS)
    usage = RankFeatures()
    last_updated_timestamp = Date()
    resource_type = Keyword(required=True)

    class Meta:
        meta = MetaField({'version': 2})


class Table(SearchableResource):
    # overwrite table name because it requires a different ngram subfield
    name = Text(required=True,
                fields={
                    "keyword": Subfield.keyword,
                    "general": Subfield.general,
                    "ngram": Subfield.get_ngram_subfield(
                        field_name="table_name",
                        max_shingle_size=8,
                        token_separator="_"
                    )
                },
                analyzer=Analyzer.stemming_analyzer,
                term_vector=POSITIONS_OFFSETS)
    columns = Text(multi=True,
                   fields={
                       "keyword": Subfield.keyword,
                       "general": Subfield.general_multi,
                       "ngram": Subfield.get_ngram_subfield(
                           field_name="table_columns",
                           multi=True,
                           token_separator="_")
                   },
                   term_vector=POSITIONS_OFFSETS,
                   analyzer=Analyzer.stemming_analyzer)
    display_name = Text(required=True,
                        fields={"keyword": Subfield.keyword},
                        analyzer=Analyzer.general_analyzer,
                        term_vector=POSITIONS_OFFSETS)
    database = Text(required=True,
                    fields={
                        "keyword": Subfield.keyword
                    },
                    analyzer=Analyzer.general_analyzer,
                    term_vector=POSITIONS_OFFSETS)
    cluster = Text(required=True,
                   fields={
                       "keyword": Subfield.keyword
                   },
                   analyzer=Analyzer.general_analyzer,
                   term_vector=POSITIONS_OFFSETS)
    schema = Text(required=True,
                  fields={
                      "keyword": Subfield.keyword,
                      "general": Subfield.general
                  },
                  analyzer=Analyzer.stemming_analyzer,
                  term_vector=POSITIONS_OFFSETS)
    column_descriptions = Text(multi=True,
                               fields={
                                   "alphanumeric": Subfield.alphanumeric_multi,
                                   "general": Subfield.general_multi
                               },
                               analyzer=Analyzer.english_analyzer,
                               term_vector=POSITIONS_OFFSETS)


class Dashboard(SearchableResource):
    group_name = Text(required=True,
                      fields={
                          "keyword": Subfield.keyword,
                          "general": Subfield.general
                      },
                      analyzer=Analyzer.stemming_analyzer,
                      term_vector=POSITIONS_OFFSETS)
    group_description = Text(analyzer=Analyzer.english_analyzer,
                             term_vector=POSITIONS_OFFSETS)
    query_names = Text(multi=True,
                       fields={
                           "keyword": Subfield.keyword,
                           "general": Subfield.general_multi,
                           "ngram": Subfield.get_ngram_subfield(
                               field_name="dashboard_query",
                               multi=True
                           )
                       },
                       analyzer=Analyzer.stemming_analyzer,
                       term_vector=POSITIONS_OFFSETS)
    chart_names = Text(multi=True,
                       fields={
                           "keyword": Subfield.keyword,
                           "general": Subfield.general_multi,
                           "ngram": Subfield.get_ngram_subfield(
                               field_name="dashboard_chart",
                               multi=True
                           )
                       },
                       analyzer=Analyzer.stemming_analyzer,
                       term_vector=POSITIONS_OFFSETS)


class Feature(SearchableResource):
    feature_group = Text(required=True,
                         fields={
                             "keyword": Subfield.keyword,
                             "general": Subfield.general_multi
                         },
                         analyzer=Analyzer.stemming_analyzer,
                         term_vector=POSITIONS_OFFSETS)
    version = Keyword(required=True)
    status = Keyword()
    entity = Text(multi=True,
                  fields={"keyword": Subfield.keyword},
                  analyzer=Analyzer.general_analyzer,
                  term_vector=POSITIONS_OFFSETS)
    availability = Keyword()


class User(SearchableResource):
    # key is email
    # name is full name, no separate first and last name
    # total read, total own, total follow goes under usage metrics
    first_name = Text(required=True,
                      fields={
                          "keyword": Keyword(),
                          "general": Subfield.general
                      },
                      analyzer=Analyzer.stemming_analyzer,
                      term_vector=POSITIONS_OFFSETS)
    last_name = Text(required=True,
                     fields={
                         "keyword": Keyword(),
                         "general": Subfield.general
                     },
                     analyzer=Analyzer.stemming_analyzer,
                     term_vector=POSITIONS_OFFSETS)


RESOURCE_TO_MAPPING: Dict[str, Document] = {
    'table': Table,
    'dashboard': Dashboard,
    'feature': Feature,
    'user': User,
    'base': SearchableResource,
}



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
from typing import (
    Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import (
    AtlasCommonParams, AtlasCommonTypes, AtlasTableKey,
)
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardUsage as RDSDashboardUsage
from amundsen_rds.models.table import TableUsage as RDSTableUsage
from amundsen_rds.models.user import User as RDSUser

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.usage.usage_constants import (
    READ_RELATION_COUNT_PROPERTY, READ_RELATION_TYPE, READ_REVERSE_RELATION_TYPE,
)
from databuilder.models.user import User
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


class Usage(GraphSerializable, TableSerializable, AtlasSerializable):
    LABELS_PERMITTED_TO_HAVE_USAGE = ['Table', 'Dashboard', 'Feature']

    def __init__(self,
                 start_label: str,
                 start_key: str,
                 user_email: str,
                 read_count: int = 1) -> None:

        if start_label not in Usage.LABELS_PERMITTED_TO_HAVE_USAGE:
            raise Exception(f'usage for {start_label} is not supported')

        self.start_label = start_label
        self.start_key = start_key
        self.user_email = user_email.strip().lower()
        self.read_count = int(read_count)

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def __repr__(self) -> str:
        return f"Usage(start_label={self.start_label!r}, start_key={self.start_key!r}, " \
               f"user_email={self.user_email!r}, read_count={self.read_count!r})"

    def create_next_node(self) -> Optional[GraphNode]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Optional[RDSModel]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        if self.user_email:
            yield GraphNode(
                key=User.get_user_model_key(email=self.user_email),
                label=User.USER_NODE_LABEL,
                attributes={
                    User.USER_NODE_EMAIL: self.user_email,
                }
            )

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=self.start_label,
            start_key=self.start_key,
            end_label=User.USER_NODE_LABEL,
            end_key=User.get_user_model_key(email=self.user_email),
            type=READ_REVERSE_RELATION_TYPE,
            reverse_type=READ_RELATION_TYPE,
            attributes={
                READ_RELATION_COUNT_PROPERTY: self.read_count,
            }
        )

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        if self.user_email:
            yield RDSUser(
                rk=User.get_user_model_key(email=self.user_email),
                email=self.user_email
            )

        if self.start_label == TableMetadata.TABLE_NODE_LABEL:
            yield RDSTableUsage(user_rk=User.get_user_model_key(email=self.user_email),
                                table_rk=self.start_key,
                                read_count=self.read_count)
        elif self.start_label == DashboardMetadata.DASHBOARD_NODE_LABEL:
            yield RDSDashboardUsage(
                user_rk=User.get_user_model_key(email=self.user_email),
                dashboard_rk=self.start_key,
                read_count=self.read_count,
            )
        else:
            raise Exception(f'{self.start_label} usage is not table serializable')

    def _get_user_key(self) -> str:
        return User.get_user_model_key(email=self.user_email)

    def _get_reader_key(self) -> str:
        return f'{self.start_key}/_reader/{self._get_user_key()}'

    def _get_entity_type(self) -> str:
        if self.start_label == 'Table':
            entity_type = AtlasTableKey(self.start_key).entity_type
        else:
            entity_type = self.start_label

        return entity_type

    def _create_atlas_user_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_user_key()),
            ('email', self._get_user_key())
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.user,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )
        return entity

    def _create_atlas_reader_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_reader_key()),
            ('count', self.read_count),
            ('entityUri', self.start_label)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.reader,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )
        return entity

    def _create_atlas_reader_dataset_relation(self) -> AtlasRelationship:
        relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.referenceable_reader,
            entityType1=self._get_entity_type(),
            entityQualifiedName1=self.start_key,
            entityType2=AtlasCommonTypes.reader,
            entityQualifiedName2=self._get_reader_key(),
            attributes=dict(count=self.read_count)
        )
        return relationship

    def _create_atlas_user_reader_relation(self) -> AtlasRelationship:
        relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.reader_user,
            entityType1=AtlasCommonTypes.reader,
            entityQualifiedName1=self._get_reader_key(),
            entityType2=AtlasCommonTypes.user,
            entityQualifiedName2=self._get_user_key(),
            attributes={}
        )
        return relationship

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        yield self._create_atlas_user_entity()
        yield self._create_atlas_reader_entity()

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)  # type: ignore
        except StopIteration:
            return None

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)  # type: ignore
        except StopIteration:
            return None

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        yield self._create_atlas_reader_dataset_relation()
        yield self._create_atlas_user_reader_relation()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Tuple

from pyhocon import ConfigTree

from databuilder.extractor.bigquery_usage_extractor import TableColumnUsageTuple
from databuilder.models.table_column_usage import ColumnReader, TableColumnUsage
from databuilder.transformer.base_transformer import Transformer


class BigqueryUsageTransformer(Transformer):

    def init(self, conf: ConfigTree) -> None:
        """
        Transformer to convert TableColumnUsageTuple data to bigquery usage data
        which can be uploaded to Neo4j
        """
        self.conf = conf

    def transform(self, record: Tuple[TableColumnUsageTuple, int]) -> Optional[TableColumnUsage]:
        if not record:
            return None

        (key, count) = record

        if not isinstance(key, TableColumnUsageTuple):
            raise Exception("BigqueryUsageTransformer expects record of type TableColumnUsageTuple")

        col_readers = []
        col_readers.append(ColumnReader(database=key.database,
                                        cluster=key.cluster,
                                        schema=key.schema,
                                        table=key.table,
                                        column=key.column,
                                        user_email=key.email,
                                        read_count=count))

        return TableColumnUsage(col_readers=col_readers)

    def get_scope(self) -> str:
        return 'transformer.bigquery_usage'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

SCHEMA_NODE_LABEL = 'Schema'

SCHEMA_NAME_ATTR = 'name'

SCHEMA_RELATION_TYPE = 'SCHEMA'
SCHEMA_REVERSE_RELATION_TYPE = 'SCHEMA_OF'

DATABASE_SCHEMA_KEY_FORMAT = '{db}://{cluster}.{schema}'

# pattern used to match a schema key, e.g., hive://gold.test_schema
SCHEMA_KEY_PATTERN_REGEX = '([a-zA-Z0-9_]+://[a-zA-Z0-9_-]+).[a-zA-Z0-9_.-]+'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

class AtlasRelationshipTypes:
    group_dashboard = 'DashboardGroup__Dashboard'
    resource_owner = 'DataSet_Users_Owner'
    dashboard_query = 'Dashboard__DashboardQuery'
    query_chart = 'DashboardQuery__DashboardChart'
    dashboard_execution = 'Dashboard__DashboardExecution'
    table_dashboard = 'Table__Dashboard'
    dashboard_owner = 'Dashboard_Users_Owner'
    table_application = 'DataSet__Application'
    table_source = 'Source__Tables'
    lineage_upstream = 'dataset_process_inputs'
    lineage_downstream = 'process_dataset_outputs'
    referenceable_reader = 'Referenceable_Readers'
    reader_user = 'Reader_Users'
    referenceable_report = 'Referenceable__Report'

    # These are just `virtual` relationship types which do not actually exist in Atlas.
    # We use those constant values to distinguish Atlas Python Client methods which should be used for populating
    # such data.
    # Tags are published using Glossary API, badges using Classification API. Other entities are published using regular
    # Entity API.
    tag = 'AtlasGlossaryTerm'
    badge = 'AtlasClassification'


class AtlasSerializedRelationshipFields:
    relation_type = 'relationshipType'
    entity_type_1 = 'entityType1'
    qualified_name_1 = 'entityQualifiedName1'
    entity_type_2 = 'entityType2'
    qualified_name_2 = 'entityQualifiedName2'


class AtlasSerializedEntityFields:
    operation = 'operation'
    relationships = 'relationships'
    type_name = 'typeName'
    attributes = 'attributes'
    relationships_separator = '|'
    relationships_kv_separator = '#'


class AtlasSerializedEntityOperation:
    CREATE = 'CREATE'
    UPDATE = 'UPDATE'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
import logging
from typing import Any, Dict

from pyhocon import ConfigTree

from databuilder.transformer.base_transformer import Transformer

MODEL_CLASS = 'model_class'

LOGGER = logging.getLogger(__name__)


class DictToModel(Transformer):
    """
    Transforms dictionary into model
    """

    def init(self, conf: ConfigTree) -> None:
        model_class = conf.get_string(MODEL_CLASS)
        module_name, class_name = model_class.rsplit(".", 1)
        mod = importlib.import_module(module_name)
        self._model_class = getattr(mod, class_name)

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        return self._model_class(**record)

    def get_scope(self) -> str:
        return 'transformer.dict_to_model'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import csv
import logging
import os
import shutil
from csv import DictWriter
from typing import (
    Any, Dict, FrozenSet,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder.job.base_job import Job
from databuilder.loader.base_loader import Loader
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.serializers import neo4_serializer
from databuilder.utils.closer import Closer

LOGGER = logging.getLogger(__name__)


class FsNeo4jCSVLoader(Loader):
    """
    Write node and relationship CSV file(s) that can be consumed by
    Neo4jCsvPublisher.
    It assumes that the record it consumes is instance of Neo4jCsvSerializable
    """
    # Config keys
    NODE_DIR_PATH = 'node_dir_path'
    RELATION_DIR_PATH = 'relationship_dir_path'
    FORCE_CREATE_DIR = 'force_create_directory'
    SHOULD_DELETE_CREATED_DIR = 'delete_created_directories'

    _DEFAULT_CONFIG = ConfigFactory.from_dict({
        SHOULD_DELETE_CREATED_DIR: True,
        FORCE_CREATE_DIR: False
    })

    def __init__(self) -> None:
        self._node_file_mapping: Dict[Any, DictWriter] = {}
        self._relation_file_mapping: Dict[Any, DictWriter] = {}
        self._keys: Dict[FrozenSet[str], int] = {}
        self._closer = Closer()

    def init(self, conf: ConfigTree) -> None:
        """
        Initializing FsNeo4jCsvLoader by creating directory for node files
        and relationship files. Note that the directory defined in
        configuration should not exist.
        :param conf:
        :return:
        """
        conf = conf.with_fallback(FsNeo4jCSVLoader._DEFAULT_CONFIG)

        self._node_dir = conf.get_string(FsNeo4jCSVLoader.NODE_DIR_PATH)
        self._relation_dir = \
            conf.get_string(FsNeo4jCSVLoader.RELATION_DIR_PATH)

        self._delete_created_dir = \
            conf.get_bool(FsNeo4jCSVLoader.SHOULD_DELETE_CREATED_DIR)
        self._force_create_dir = conf.get_bool(FsNeo4jCSVLoader.FORCE_CREATE_DIR)
        self._create_directory(self._node_dir)
        self._create_directory(self._relation_dir)

    def _create_directory(self, path: str) -> None:
        """
        Validate directory does not exist, creates it, register deletion of
        created directory function to Job.closer.
        :param path:
        :return:
        """
        if os.path.exists(path):
            if self._force_create_dir:
                LOGGER.info('Directory exist. Deleting directory %s', path)
                shutil.rmtree(path)
            else:
                raise RuntimeError(f'Directory should not exist: {path}')

        os.makedirs(path)

        def _delete_dir() -> None:
            if not self._delete_created_dir:
                LOGGER.warning('Skip Deleting directory %s', path)
                return

            LOGGER.info('Deleting directory %s', path)
            shutil.rmtree(path)

        # Directory should be deleted after publish is finished
        Job.closer.register(_delete_dir)

    def load(self, csv_serializable: GraphSerializable) -> None:
        """
        Writes Neo4jCsvSerializable into CSV files.
        There are multiple CSV files that this method writes.
        This is because there're not only node and relationship, but also it
        can also have different nodes, and relationships.

        Common pattern for both nodes and relations:
         1. retrieve csv row (a dict where keys represent a header,
         values represent a row)
         2. using this dict to get a appropriate csv writer and write to it.
         3. repeat 1 and 2

        :param csv_serializable:
        :return:
        """

        node = csv_serializable.next_node()
        while node:
            node_dict = neo4_serializer.serialize_node(node)
            key = (node.label, self._make_key(node_dict))
            file_suffix = '{}_{}'.format(*key)
            node_writer = self._get_writer(node_dict,
                                           self._node_file_mapping,
                                           key,
                                           self._node_dir,
                                           file_suffix)
            node_writer.writerow(node_dict)
            node = csv_serializable.next_node()

        relation = csv_serializable.next_relation()
        while relation:
            relation_dict = neo4_serializer.serialize_relationship(relation)
            key2 = (relation.start_label,
                    relation.end_label,
                    relation.type,
                    self._make_key(relation_dict))

            file_suffix = f'{key2[0]}_{key2[1]}_{key2[2]}'
            relation_writer = self._get_writer(relation_dict,
                                               self._relation_file_mapping,
                                               key2,
                                               self._relation_dir,
                                               file_suffix)
            relation_writer.writerow(relation_dict)
            relation = csv_serializable.next_relation()

    def _get_writer(self,
                    csv_record_dict: Dict[str, Any],
                    file_mapping: Dict[Any, DictWriter],
                    key: Any,
                    dir_path: str,
                    file_suffix: str
                    ) -> DictWriter:
        """
        Finds a writer based on csv record, key.
        If writer does not exist, it's creates a csv writer and update the
        mapping.

        :param csv_record_dict:
        :param file_mapping:
        :param key:
        :param file_suffix:
        :return:
        """
        writer = file_mapping.get(key)
        if writer:
            return writer

        LOGGER.info('Creating file for %s', key)

        file_out = open(f'{dir_path}/{file_suffix}.csv', 'w', encoding='utf8')
        writer = csv.DictWriter(file_out, fieldnames=csv_record_dict.keys(),
                                quoting=csv.QUOTE_NONNUMERIC)

        def file_out_close() -> None:
            LOGGER.info('Closing file IO %s', file_out)
            file_out.close()
        self._closer.register(file_out_close)

        writer.writeheader()
        file_mapping[key] = writer

        return writer

    def close(self) -> None:
        """
        Any closeable callable registered in _closer, it will close.
        :return:
        """
        self._closer.close()

    def get_scope(self) -> str:
        return "loader.filesystem_csv_neo4j"

    def _make_key(self, record_dict: Dict[str, Any]) -> int:
        """ Each unique set of record keys is assigned an increasing numeric key """
        return self._keys.setdefault(frozenset(record_dict.keys()), len(self._keys))



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
from collections import namedtuple

AtlasRelationship = namedtuple(
    'AtlasRelationship',
    [
        'relationshipType',
        'entityType1',
        'entityQualifiedName1',
        'entityType2',
        'entityQualifiedName2',
        'attributes',
    ],
)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from amundsen_common.models.index_map import (
    DASHBOARD_ELASTICSEARCH_INDEX_MAPPING as DASHBOARD_INDEX_MAP, TABLE_INDEX_MAP, USER_INDEX_MAP,
)

# Please use constants in amundsen_common.models.index_map directly. This file is only here
# for backwards compatibility.
TABLE_ELASTICSEARCH_INDEX_MAPPING = TABLE_INDEX_MAP
DASHBOARD_ELASTICSEARCH_INDEX_MAPPING = DASHBOARD_INDEX_MAP
USER_ELASTICSEARCH_INDEX_MAPPING = USER_INDEX_MAP



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterable, Iterator, Optional, Union,
)

from amundsen_rds.models import RDSModel

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.usage.usage import Usage


class ColumnReader(Usage):
    """
    Represent user's read action on a table - and eventually on a column.
    """

    def __init__(self,
                 database: str,
                 cluster: str,
                 schema: str,
                 table: str,
                 column: str,  # not used: per-column usage not yet implemented
                 user_email: str,
                 read_count: int = 1
                 ) -> None:

        Usage.__init__(
            self,
            start_label=TableMetadata.TABLE_NODE_LABEL,
            start_key=TableMetadata.TABLE_KEY_FORMAT.format(
                db=database,
                cluster=cluster,
                schema=schema,
                tbl=table),
            user_email=user_email,
            read_count=read_count,
        )


class TableColumnUsage(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    Represents an iterable of read actions.
    """

    def __init__(self, col_readers: Iterable[ColumnReader]) -> None:
        self.col_readers = col_readers

        self._node_iterator = self._create_node_iterator()
        self._rel_iter = self._create_rel_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        for usage in self.col_readers:
            node = usage.create_next_node()
            while node is not None:
                yield node
                node = usage.create_next_node()

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    def _create_rel_iterator(self) -> Iterator[GraphRelationship]:
        for usage in self.col_readers:
            rel = usage.create_next_relation()
            while rel is not None:
                yield rel
                rel = usage.create_next_relation()

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        for usage in self.col_readers:
            record = usage.create_next_record()
            while record is not None:
                yield record
                record = usage.create_next_record()

    def _create_next_atlas_entity(self) -> Iterator[Optional[AtlasEntity]]:
        for usage in self.col_readers:
            yield usage.create_next_atlas_entity()

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)  # type: ignore
        except StopIteration:
            return None

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)  # type: ignore
        except StopIteration:
            return None

    def _create_atlas_relation_iterator(self) -> Iterator[Optional[AtlasRelationship]]:
        for usage in self.col_readers:
            yield usage.create_next_atlas_relation()

    def __repr__(self) -> str:
        return f'TableColumnUsage(col_readers={self.col_readers!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasDashboardTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardTimestamp as RDSDashboardTimestamp

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.timestamp import timestamp_constants
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasSerializedEntityOperation

LOGGER = logging.getLogger(__name__)


class DashboardLastModifiedTimestamp(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    A model that encapsulate Dashboard's last modified timestamp in epoch
    """

    DASHBOARD_LAST_MODIFIED_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group_id}/' \
                                         '{dashboard_id}/_last_modified_timestamp'

    def __init__(self,
                 dashboard_group_id: Optional[str],
                 dashboard_id: Optional[str],
                 last_modified_timestamp: int,
                 product: Optional[str] = '',
                 cluster: str = 'gold',
                 **kwargs: Any
                 ) -> None:
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        self._last_modified_timestamp = last_modified_timestamp
        self._product = product
        self._cluster = cluster
        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()
        self._record_iterator = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes = {
            timestamp_constants.TIMESTAMP_PROPERTY: self._last_modified_timestamp,
            timestamp_constants.TIMESTAMP_NAME_PROPERTY: timestamp_constants.TimestampName.last_updated_timestamp.name
        }
        node = GraphNode(
            key=self._get_last_modified_node_key(),
            label=timestamp_constants.NODE_LABEL,
            attributes=node_attributes
        )
        yield node

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relationship = GraphRelationship(
            start_key=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            ),
            start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            end_key=self._get_last_modified_node_key(),
            end_label=timestamp_constants.NODE_LABEL,
            type=timestamp_constants.LASTUPDATED_RELATION_TYPE,
            reverse_type=timestamp_constants.LASTUPDATED_REVERSE_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:

        # last modified
        attrs_mapping = [
            (
                AtlasCommonParams.qualified_name, DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                    product=self._product,
                    cluster=self._cluster,
                    dashboard_group=self._dashboard_group_id,
                    dashboard_name=self._dashboard_id
                )
            ),
            (AtlasCommonParams.last_modified_timestamp, self._last_modified_timestamp),
        ]

        dashboard_entity_attrs = get_entity_attrs(attrs_mapping)

        last_modified = AtlasEntity(
            typeName=AtlasDashboardTypes.metadata,
            operation=AtlasSerializedEntityOperation.UPDATE,
            relationships=None,
            attributes=dashboard_entity_attrs
        )
        yield last_modified

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        yield RDSDashboardTimestamp(
            rk=self._get_last_modified_node_key(),
            timestamp=self._last_modified_timestamp,
            name=timestamp_constants.TimestampName.last_updated_timestamp.name,
            dashboard_rk=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            )
        )

    def _get_last_modified_node_key(self) -> str:
        return DashboardLastModifiedTimestamp.DASHBOARD_LAST_MODIFIED_KEY_FORMAT.format(
            product=self._product,
            cluster=self._cluster,
            dashboard_group_id=self._dashboard_group_id,
            dashboard_id=self._dashboard_id,
        )

    def __repr__(self) -> str:
        return f'DashboardLastModifiedTimestamp({self._dashboard_group_id!r}, {self._dashboard_id!r}, ' \
               f'{self._last_modified_timestamp!r}, {self._product!r}, {self._cluster!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, List

from pyhocon import ConfigFactory, ConfigTree

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.transformer.base_transformer import ChainedTransformer, Transformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel
from databuilder.transformer.timestamp_string_to_epoch import FIELD_NAME, TimestampStringToEpoch

LOGGER = logging.getLogger(__name__)


class ModeDashboardExecutionsExtractor(Extractor):
    """
    A Extractor that extracts run (execution) status and timestamp.

    """

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(
            restapi_query=restapi_query,
            conf=self._conf
        )

        # Payload from RestApiQuery has timestamp which is ISO8601. Here we are using TimestampStringToEpoch to
        # transform into epoch and then using DictToModel to convert Dictionary to Model
        transformers: List[Transformer] = []
        timestamp_str_to_epoch_transformer = TimestampStringToEpoch()
        timestamp_str_to_epoch_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, timestamp_str_to_epoch_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict({FIELD_NAME: 'execution_timestamp', })))

        transformers.append(timestamp_str_to_epoch_transformer)

        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_execution.DashboardExecution'})))
        transformers.append(dict_to_model_transformer)

        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self) -> Any:
        record = self._extractor.extract()
        if not record:
            return None

        return self._transformer.transform(record=record)

    def get_scope(self) -> str:
        return 'extractor.mode_dashboard_execution'

    def _build_restapi_query(self) -> ModePaginatedRestApiQuery:
        """
        Build REST API Query to get Mode Dashboard last execution.
        :return: A RestApiQuery that provides Mode Dashboard execution (run)
        """

        seed_query = ModeDashboardUtils.get_seed_query(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf, discover_auth=True)

        # Reports
        # https://mode.com/developer/discovery-api/analytics/reports/
        url = 'https://app.mode.com/batch/{organization}/reports'
        json_path = 'reports[*].[token, space_token, last_run_at, last_run_state]'
        field_names = ['dashboard_id', 'dashboard_group_id', 'execution_timestamp', 'execution_state']
        max_record_size = 1000
        pagination_json_path = 'reports[*]'
        last_execution_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=url, params=params,
                                                         json_path=json_path, field_names=field_names,
                                                         skip_no_result=True, max_record_size=max_record_size,
                                                         pagination_json_path=pagination_json_path)

        return last_execution_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import (
    Any, Dict, List, Optional,
)

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship

NEPTUNE_HEADER_ID = "~id"
NEPTUNE_HEADER_LABEL = "~label"
NEPTUNE_RELATIONSHIP_HEADER_FROM = "~from"
NEPTUNE_RELATIONSHIP_HEADER_TO = "~to"


METADATA_KEY_PROPERTY_NAME = 'key'
METADATA_KEY_PROPERTY_NAME_BULK_LOADER_FORMAT = '{name}:String(single)'.format(
    name=METADATA_KEY_PROPERTY_NAME
)

# last seen property names
NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME = "last_extracted_datetime"
NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT = "{name}:Date(single)".format(
    name=NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME
)
NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME = "last_extracted_datetime"
NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT = "{name}:Date(single)".format(
    name=NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME
)
# Add a property showing where the the node or relationship comes from
NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME = "creation_type"
NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT = "{name}:String(single)".format(
    name=NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME
)
NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME = "creation_type"
NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT = "{name}:String".format(
    name=NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME
)

NEPTUNE_CREATION_TYPE_JOB = "job"


def convert_relationship(relationship: Optional[GraphRelationship]) -> List[Dict[str, Any]]:
    if relationship is None:
        return []

    if relationship.start_key == '' or relationship.end_key == '':
        return []

    neptune_start_key = "{label}:{key}".format(
        label=relationship.start_label,
        key=relationship.start_key
    )
    neptune_end_key = "{label}:{key}".format(
        label=relationship.end_label,
        key=relationship.end_key
    )
    relation_id = get_forward_relationship_id(relationship)
    relation_id_reverse = get_reverse_relationship_id(relationship)
    current_string_time = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')
    forward_relationship_doc = {
        NEPTUNE_HEADER_ID: relation_id,
        METADATA_KEY_PROPERTY_NAME_BULK_LOADER_FORMAT: relation_id,
        NEPTUNE_RELATIONSHIP_HEADER_FROM: neptune_start_key,
        NEPTUNE_RELATIONSHIP_HEADER_TO: neptune_end_key,
        NEPTUNE_HEADER_LABEL: relationship.type,
        NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: current_string_time,
        NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,

    }

    reverse_relationship_doc = {
        NEPTUNE_HEADER_ID: relation_id_reverse,
        METADATA_KEY_PROPERTY_NAME_BULK_LOADER_FORMAT: relation_id_reverse,
        NEPTUNE_RELATIONSHIP_HEADER_FROM: neptune_end_key,
        NEPTUNE_RELATIONSHIP_HEADER_TO: neptune_start_key,
        NEPTUNE_HEADER_LABEL: relationship.reverse_type,
        NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: current_string_time,
        NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
    }

    for key, value in relationship.attributes.items():
        neptune_value_type = _get_neptune_type_for_value(value)
        doc_key = "{key_name}:{neptune_value_type}(single)".format(
            key_name=key,
            neptune_value_type=neptune_value_type
        )
        forward_relationship_doc[doc_key] = value
        reverse_relationship_doc[doc_key] = value

    return [
        forward_relationship_doc,
        reverse_relationship_doc
    ]


def get_forward_relationship_id(relationship: GraphRelationship) -> str:
    return "{label}:{from_vertex_label}:{from_vertex_id}_{to_vertex_label}:{to_vertex_id}".format(
        from_vertex_id=relationship.start_key,
        from_vertex_label=relationship.start_label,
        to_vertex_id=relationship.end_key,
        to_vertex_label=relationship.end_label,
        label=relationship.type
    )


def get_reverse_relationship_id(relationship: GraphRelationship) -> str:
    return "{label}:{from_vertex_label}:{from_vertex_id}_{to_vertex_label}:{to_vertex_id}".format(
        to_vertex_id=relationship.start_key,
        to_vertex_label=relationship.start_label,
        from_vertex_id=relationship.end_key,
        from_vertex_label=relationship.end_label,
        label=relationship.reverse_type
    )


def convert_node(node: Optional[GraphNode]) -> Dict[str, Any]:
    if node is None:
        return {}

    if not node.key:
        return {}

    current_string_time = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')
    node_id = get_node_id(node)
    node_dict = {
        NEPTUNE_HEADER_ID: node_id,
        METADATA_KEY_PROPERTY_NAME_BULK_LOADER_FORMAT: node.key,
        NEPTUNE_HEADER_LABEL: node.label,
        NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: current_string_time,
        NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
    }

    for attr_key, attr_value in node.attributes.items():
        neptune_value_type = _get_neptune_type_for_value(attr_value)
        doc_key = "{key_name}:{neptune_value_type}(single)".format(
            key_name=attr_key,
            neptune_value_type=neptune_value_type
        )
        if doc_key not in node_dict:
            node_dict[doc_key] = attr_value

    return node_dict


def get_node_id(node: GraphNode) -> str:
    return "{label}:{key}".format(
        label=node.label,
        key=node.key
    )


def _get_neptune_type_for_value(value: Any) -> Optional[str]:
    if isinstance(value, str):
        return "String"
    elif isinstance(value, bool):
        return "Bool"
    elif isinstance(value, int):
        return "Long"
    elif isinstance(value, float):
        return "Double"

    return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator, Union

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasCommonTypes

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


class ResourceReport(GraphSerializable, AtlasSerializable):
    """
    Resource Report matching model

    Report represents a document that can be linked to any resource (like a table) in Amundsen.

    Example would be Pandas Profiling HTML report containing full advanced profile of a table.
    """

    RESOURCE_REPORT_LABEL = 'Report'

    RESOURCE_REPORT_NAME = 'name'
    RESOURCE_REPORT_URL = 'url'

    REPORT_KEY_FORMAT = '{resource_uri}/_report/{report_name}'

    REPORT_RESOURCE_RELATION_TYPE = 'REFERS_TO'
    RESOURCE_REPORT_RELATION_TYPE = 'HAS_REPORT'

    def __init__(self,
                 name: str,
                 url: str,
                 resource_uri: str,
                 resource_label: str,  # for example 'Table'
                 ) -> None:
        self.report_name = name
        self.report_url = url

        self.resource_uri = resource_uri
        self.resource_label = resource_label

        self.resource_report_key = self.get_resource_model_key()

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def get_resource_model_key(self) -> str:
        return ResourceReport.REPORT_KEY_FORMAT.format(resource_uri=self.resource_uri, report_name=self.report_name)

    def create_next_node(self) -> Union[GraphNode, None]:
        # creates new node
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create an application node
        :return:
        """
        report_node = GraphNode(
            key=self.resource_report_key,
            label=ResourceReport.RESOURCE_REPORT_LABEL,
            attributes={
                ResourceReport.RESOURCE_REPORT_NAME: self.report_name,
                ResourceReport.RESOURCE_REPORT_URL: self.report_url
            }
        )

        yield report_node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relations between application and table nodes
        :return:
        """
        graph_relationship = GraphRelationship(
            start_key=self.resource_uri,
            start_label=self.resource_label,
            end_key=self.resource_report_key,
            end_label=ResourceReport.RESOURCE_REPORT_LABEL,
            type=ResourceReport.RESOURCE_REPORT_RELATION_TYPE,
            reverse_type=ResourceReport.REPORT_RESOURCE_RELATION_TYPE,
            attributes={}
        )

        yield graph_relationship

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        group_attrs_mapping = [
            (AtlasCommonParams.qualified_name, self.resource_report_key),
            ('name', self.report_name),
            ('url', self.report_url)
        ]

        entity_attrs = get_entity_attrs(group_attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.resource_report,
            operation=AtlasSerializedEntityOperation.CREATE,
            relationships=None,
            attributes=entity_attrs,
        )

        yield entity

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.referenceable_report,
            entityType1=self.resource_label,
            entityQualifiedName1=self.resource_uri,
            entityType2=AtlasCommonTypes.resource_report,
            entityQualifiedName2=self.resource_report_key,
            attributes={}
        )

        yield relationship



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

GraphNode = namedtuple(
    'GraphNode',
    [
        'key',
        'label',
        'attributes'
    ]
)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import logging
from os import listdir
from os.path import isfile, join
from typing import (
    Any, Dict, Iterator, List, Tuple,
)

import pandas
from amundsen_common.utils.atlas import AtlasCommonParams, AtlasCommonTypes
from apache_atlas.exceptions import AtlasServiceException
from apache_atlas.model.glossary import (
    AtlasGlossary, AtlasGlossaryHeader, AtlasGlossaryTerm,
)
from apache_atlas.model.instance import (
    AtlasEntitiesWithExtInfo, AtlasEntity, AtlasObjectId, AtlasRelatedObjectId,
)
from apache_atlas.model.relationship import AtlasRelationship
from apache_atlas.model.typedef import AtlasClassificationDef, AtlasTypesDef
from pyhocon import ConfigTree

from databuilder.publisher.base_publisher import Publisher
from databuilder.types.atlas import AtlasEntityInitializer
from databuilder.utils.atlas import (
    AtlasRelationshipTypes, AtlasSerializedEntityFields, AtlasSerializedEntityOperation,
    AtlasSerializedRelationshipFields,
)

LOGGER = logging.getLogger(__name__)


class AtlasCSVPublisher(Publisher):
    # atlas client
    ATLAS_CLIENT = 'atlas_client'
    # A directory that contains CSV files for entities
    ENTITY_DIR_PATH = 'entity_files_directory'
    # A directory that contains CSV files for relationships
    RELATIONSHIP_DIR_PATH = 'relationship_files_directory'
    # atlas create entity batch size
    ATLAS_ENTITY_CREATE_BATCH_SIZE = 'batch_size'
    # whether entity types should be registered before data is synced to Atlas
    REGISTER_ENTITY_TYPES = 'register_entity_types'

    def __init__(self) -> None:
        super().__init__()

    def init(self, conf: ConfigTree) -> None:
        self._entity_files = self._list_files(conf, AtlasCSVPublisher.ENTITY_DIR_PATH)
        self._relationship_files = self._list_files(conf, AtlasCSVPublisher.RELATIONSHIP_DIR_PATH)
        self._config = conf
        self._atlas_client = self._config.get(AtlasCSVPublisher.ATLAS_CLIENT)
        self._register_entity_types = self._config.get_bool(AtlasCSVPublisher.REGISTER_ENTITY_TYPES, True)

        if self._register_entity_types:
            LOGGER.info('Registering Atlas Entity Types.')

            try:
                init = AtlasEntityInitializer(self._atlas_client)
                init.create_required_entities()

                LOGGER.info('Registered Atlas Entity Types.')
            except Exception:
                LOGGER.error('Failed to register Atlas Entity Types.', exc_info=True)

    def _list_files(self, conf: ConfigTree, path_key: str) -> List[str]:
        """
        List files from directory
        :param conf:
        :param path_key:
        :return: List of file paths
        """
        if path_key not in conf:
            return []

        path = conf.get_string(path_key)
        return sorted(join(path, f) for f in listdir(path) if isfile(join(path, f)))

    def publish_impl(self) -> None:
        """
        Publishes Entities first and then Relations
        :return:
        """
        LOGGER.info('Creating entities using Entity files: %s', self._entity_files)
        for entity_file in self._entity_files:
            entities_to_create, entities_to_update, \
                glossary_terms_create, classifications_create = self._create_entity_instances(entity_file=entity_file)
            self._sync_entities_to_atlas(entities_to_create)
            self._update_entities(entities_to_update)
            self._create_glossary_terms(glossary_terms_create)
            self._create_classifications(classifications_create)

        LOGGER.info('Creating relations using relation files: %s', self._relationship_files)
        for relation_file in self._relationship_files:
            self._create_relations(relation_file=relation_file)

    def _update_entities(self, entities_to_update: List[AtlasEntity]) -> None:
        """
        Go over the entities list , create atlas relationships instances and sync them with atlas
        :param entities_to_update:
        :return:
        """
        for entity_to_update in entities_to_update:
            existing_entity = self._atlas_client.entity.get_entity_by_attribute(
                entity_to_update.attributes[AtlasCommonParams.type_name],
                [(AtlasCommonParams.qualified_name, entity_to_update.attributes[AtlasCommonParams.qualified_name])],
            )
            existing_entity.entity.attributes.update(entity_to_update.attributes)
            try:
                self._atlas_client.entity.update_entity(existing_entity)
            except AtlasServiceException:
                LOGGER.error('Fail to update entity', exc_info=True)

    def _create_relations(self, relation_file: str) -> None:
        """
        Go over the relation file, create atlas relationships instances and sync them with atlas
        :param relation_file:
        :return:
        """

        with open(relation_file, encoding='utf8') as relation_csv:
            for relation_record in pandas.read_csv(relation_csv, na_filter=False).to_dict(orient='records'):
                if relation_record[AtlasSerializedRelationshipFields.relation_type] == AtlasRelationshipTypes.tag:
                    self._assign_glossary_term(relation_record)
                    continue
                elif relation_record[AtlasSerializedRelationshipFields.relation_type] == AtlasRelationshipTypes.badge:
                    self._assign_classification(relation_record)
                    continue

                relation = self._create_relation(relation_record)
                try:
                    self._atlas_client.relationship.create_relationship(relation)
                except AtlasServiceException:
                    LOGGER.error('Fail to create atlas relationship', exc_info=True)
                except Exception as e:
                    LOGGER.error(e)

    def _render_unique_attributes(self, entity_type: str, qualified_name: str) -> Dict[Any, Any]:
        """
        Render uniqueAttributes dict, this struct is needed to identify AtlasObjects
        :param entity_type:
        :param qualified_name:
        :return: rendered uniqueAttributes dict
        """
        return {
            AtlasCommonParams.type_name: entity_type,
            AtlasCommonParams.unique_attributes: {
                AtlasCommonParams.qualified_name: qualified_name,
            },
        }

    def _get_atlas_related_object_id_by_qn(self, entity_type: str, qn: str) -> AtlasRelatedObjectId:
        return AtlasRelatedObjectId(attrs=self._render_unique_attributes(entity_type, qn))

    def _get_atlas_object_id_by_qn(self, entity_type: str, qn: str) -> AtlasObjectId:
        return AtlasObjectId(attrs=self._render_unique_attributes(entity_type, qn))

    def _create_relation(self, relation_dict: Dict[str, str]) -> AtlasRelationship:
        """
        Go over the relation dictionary file and create atlas relationships instances
        :param relation_dict:
        :return:
        """

        relation = AtlasRelationship(
            {AtlasCommonParams.type_name: relation_dict[AtlasSerializedRelationshipFields.relation_type]},
        )
        relation.end1 = self._get_atlas_object_id_by_qn(
            relation_dict[AtlasSerializedRelationshipFields.entity_type_1],
            relation_dict[AtlasSerializedRelationshipFields.qualified_name_1],
        )
        relation.end2 = self._get_atlas_object_id_by_qn(
            relation_dict[AtlasSerializedRelationshipFields.entity_type_2],
            relation_dict[AtlasSerializedRelationshipFields.qualified_name_2],
        )

        return relation

    def _create_entity_instances(self, entity_file: str) -> Tuple[List[AtlasEntity], List[AtlasEntity],
                                                                  List[Dict], List[Dict]]:
        """
        Go over the entities file and try creating instances
        :param entity_file:
        :return:
        """
        entities_to_create = []
        entities_to_update = []
        glossary_terms_to_create = []
        classifications_to_create = []
        with open(entity_file, encoding='utf8') as entity_csv:
            for entity_record in pandas.read_csv(entity_csv, na_filter=False).to_dict(orient='records'):
                if entity_record[AtlasSerializedEntityFields.type_name] == AtlasCommonTypes.tag:
                    glossary_terms_to_create.append(entity_record)
                    continue

                if entity_record[AtlasSerializedEntityFields.type_name] == AtlasCommonTypes.badge:
                    classifications_to_create.append(entity_record)
                    continue

                if entity_record[AtlasSerializedEntityFields.operation] == AtlasSerializedEntityOperation.CREATE:
                    entities_to_create.append(self._create_entity_from_dict(entity_record))
                if entity_record[AtlasSerializedEntityFields.operation] == AtlasSerializedEntityOperation.UPDATE:
                    entities_to_update.append(self._create_entity_from_dict(entity_record))
        return entities_to_create, entities_to_update, glossary_terms_to_create, classifications_to_create

    def _extract_entity_relations_details(self, relation_details: str) -> Iterator[Tuple]:
        """
        Generate relation details from relation_attr#related_entity_type#related_qualified_name
        """
        relations = relation_details.split(AtlasSerializedEntityFields.relationships_separator)
        for relation in relations:
            relation_split = relation.split(AtlasSerializedEntityFields.relationships_kv_separator)
            yield relation_split[0], relation_split[1], relation_split[2]

    def _create_entity_from_dict(self, entity_dict: Dict) -> AtlasEntity:
        """
        Create atlas entity instance from dict
        :param entity_dict:
        :return: AtlasEntity
        """
        type_name = {AtlasCommonParams.type_name: entity_dict[AtlasCommonParams.type_name]}
        entity = AtlasEntity(type_name)
        entity.attributes = entity_dict
        relationships = entity_dict.get(AtlasSerializedEntityFields.relationships)
        if relationships:
            relations = {}
            for relation_attr, rel_type, rel_qn in self._extract_entity_relations_details(relationships):
                related_obj = self._get_atlas_related_object_id_by_qn(rel_type, rel_qn)
                relations[relation_attr] = related_obj
            entity.relationshipAttributes = relations
        return entity

    def _chunks(self, lst: List) -> Iterator:
        """
        Yield successive n-sized chunks from lst.
        :param lst:
        :return: chunks generator
        """
        n = self._config.get_int(AtlasCSVPublisher.ATLAS_ENTITY_CREATE_BATCH_SIZE)
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def _sync_entities_to_atlas(self, entities: List[AtlasEntity]) -> None:
        """
        Sync entities instances with atlas
        :param entities: list of entities
        :return:
        """
        entities_chunks = self._chunks(entities)
        for entity_chunk in entities_chunks:
            LOGGER.info(f'Syncing chunk of {len(entity_chunk)} entities with atlas')
            chunk = AtlasEntitiesWithExtInfo()
            chunk.entities = entity_chunk
            try:
                self._atlas_client.entity.create_entities(chunk)
            except AtlasServiceException:
                LOGGER.error('Error during entity syncing', exc_info=True)

    def _create_glossary_terms(self, glossary_terms: List[Dict]) -> None:
        for glossary_term_spec in glossary_terms:
            glossary_name = glossary_term_spec.get('glossary')
            term_name = glossary_term_spec.get('term')

            glossary_def = AtlasGlossary({'name': glossary_name, 'shortDescription': ''})

            try:
                glossary = self._atlas_client.glossary.create_glossary(glossary_def)
            except AtlasServiceException:
                LOGGER.info(f'Glossary: {glossary_name} already exists.')
                glossary = next(filter(lambda x: x.get('name') == glossary_name,
                                       self._atlas_client.glossary.get_all_glossaries()))

            glossary_guid = glossary['guid']
            glossary_def = AtlasGlossaryHeader({'glossaryGuid': glossary_guid})
            term_def = AtlasGlossaryTerm({'name': term_name, 'anchor': glossary_def})

            try:
                self._atlas_client.glossary.create_glossary_term(term_def)
            except AtlasServiceException:
                LOGGER.info(f'Glossary Term: {term_name} already exists.')

    def _assign_glossary_term(self, relationship_spec: Dict) -> None:
        _glossary_name, _term_name = relationship_spec[AtlasSerializedRelationshipFields.qualified_name_2].split(',')

        glossary_name = _glossary_name.split('=')[1]
        term_name = _term_name.split('=')[1]

        entity_type = relationship_spec[AtlasSerializedRelationshipFields.entity_type_1]
        entity_qn = relationship_spec[AtlasSerializedRelationshipFields.qualified_name_1]

        glossary = next(filter(lambda g: g.get('name') == glossary_name,
                               self._atlas_client.glossary.get_all_glossaries()))

        glossary_guid = glossary[AtlasCommonParams.guid]

        term = next(filter(lambda t: t.get('name') == term_name,
                           self._atlas_client.glossary.get_glossary_terms(glossary_guid)))

        entity = self._atlas_client.entity.get_entity_by_attribute(entity_type, uniq_attributes=[
            (AtlasCommonParams.qualified_name, entity_qn)])

        entity_guid = entity.entity.guid

        e = AtlasRelatedObjectId({AtlasCommonParams.guid: entity_guid, AtlasCommonParams.type_name: entity_type})

        try:
            self._atlas_client.glossary.assign_term_to_entities(term[AtlasCommonParams.guid], [e])
        except Exception:
            LOGGER.error('Error assigning terms to entities.', exc_info=True)

    def _render_super_type_from_dict(self, classification_spec: Dict) -> AtlasClassificationDef:
        return self._render_classification(classification_spec, True)

    def _render_sub_type_from_dict(self, classification_spec: Dict) -> AtlasClassificationDef:
        return self._render_classification(classification_spec, False)

    def _render_classification(self, classification_spec: Dict, super_type: bool) -> AtlasClassificationDef:
        name = classification_spec.get('category') if super_type else classification_spec.get('name')
        sub_types = [classification_spec.get('category')] if not super_type else []

        result = AtlasClassificationDef(attrs=dict(name=name,
                                                   attributeDefs=[],
                                                   subTypes=sub_types,
                                                   superTypes=[],
                                                   entityTypes=[]))

        return result

    def _create_classifications(self, classifications: List[Dict]) -> None:
        _st = set()
        super_types = [self._render_super_type_from_dict(s) for s in classifications if s['category'] not in _st and
                       _st.add(s['category'])]  # type: ignore
        super_types_chunks = self._chunks(super_types)

        sub_types = [self._render_sub_type_from_dict(s) for s in classifications]
        sub_types_chunks = self._chunks(sub_types)

        for chunks in [super_types_chunks, sub_types_chunks]:
            for chunk in chunks:
                LOGGER.info(f'Syncing chunk of {len(chunk)} classifications with atlas')
                try:
                    types = AtlasTypesDef(attrs=dict(classificationDefs=chunk))

                    self._atlas_client.typedef.create_atlas_typedefs(types)
                except AtlasServiceException:
                    LOGGER.error('Error during classification syncing', exc_info=True)

    def _assign_classification(self, relationship_spec: Dict) -> None:
        classification_qn = relationship_spec[AtlasSerializedRelationshipFields.qualified_name_2]

        entity_type = relationship_spec[AtlasSerializedRelationshipFields.entity_type_1]
        entity_qn = relationship_spec[AtlasSerializedRelationshipFields.qualified_name_1]

        try:
            self._atlas_client.entity.add_classifications_by_type(entity_type,
                                                                  uniq_attributes=[(AtlasCommonParams.qualified_name,
                                                                                    entity_qn)],
                                                                  classifications=[classification_qn])
        except Exception:
            LOGGER.error('Error during classification assingment.', exc_info=True)

    def get_scope(self) -> str:
        return 'publisher.atlas_csv_publisher'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import re
from typing import (
    Any, Dict, Iterator, Optional,
)

import requests
from pyhocon import ConfigTree

import databuilder.extractor.dashboard.tableau.tableau_dashboard_constants as const
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.restapi.rest_api_extractor import STATIC_RECORD_DICT


class TableauDashboardUtils:
    """
    Provides various utility functions specifc to the Tableau dashboard extractors.
    """

    @staticmethod
    def sanitize_schema_name(schema_name: str) -> str:
        """
        Sanitizes a given string so that it can safely be used as a table's schema.
        Sanitization behaves as follows:
            - all spaces and periods are replaced by underscores
            - any [], (), -, &, and ? characters are deleted
        """
        # this indentation looks a little odd, but otherwise the linter complains
        return re.sub(r' ', '_',
                      re.sub(r'\.', '_',
                             re.sub(r'(\[|\]|\(|\)|\-|\&|\?)', '', schema_name)))

    @staticmethod
    def sanitize_database_name(database_name: str) -> str:
        """
        Sanitizes a given string so that it can safely be used as a table's database.
        Sanitization behaves as follows:
            - all hyphens are deleted
        """
        return re.sub(r'-', '', database_name)

    @staticmethod
    def sanitize_table_name(table_name: str) -> str:
        """
        Sanitizes a given string so that it can safely be used as a table name.
        Replicates the current behavior of sanitize_workbook_name, but this is purely coincidental.
        As more breaking characters/patterns are found, each method should be updated to reflect the specifics.
        Sanitization behaves as follows:
            - all forward slashes and single quotes characters are deleted
        """
        return re.sub(r'(\/|\')', '', table_name)

    @staticmethod
    def sanitize_workbook_name(workbook_name: str) -> str:
        """
        Sanitizes a given string so that it can safely be used as a workbook ID.
        Mimics the current behavior of sanitize_table_name for now, but is purely coincidental.
        As more breaking characters/patterns are found, each method should be updated to reflect the specifics.
        Sanitization behaves as follows:
            - all forward slashes and single quotes characters are deleted
        """
        return re.sub(r'(\/|\')', '', workbook_name)


class TableauGraphQLApiExtractor(Extractor):
    """
    Base class for querying the Tableau Metdata API, which uses a GraphQL schema.
    """

    API_BASE_URL = const.API_BASE_URL
    QUERY = 'query'
    QUERY_VARIABLES = 'query_variables'
    VERIFY_REQUEST = 'verify_request'

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf
        self._auth_token = TableauDashboardAuth(self._conf).token
        self._query = self._conf.get(TableauGraphQLApiExtractor.QUERY)
        self._iterator: Optional[Iterator[Dict[str, Any]]] = None
        self._static_dict = conf.get(STATIC_RECORD_DICT, dict())
        self._metadata_url = '{api_base_url}/api/metadata/graphql'.format(
            api_base_url=self._conf.get_string(TableauGraphQLApiExtractor.API_BASE_URL)
        )
        self._query_variables = self._conf.get(TableauGraphQLApiExtractor.QUERY_VARIABLES, {})
        self._verify_request = self._conf.get(TableauGraphQLApiExtractor.VERIFY_REQUEST, None)

    def execute_query(self) -> Dict[str, Any]:
        """
        Executes the extractor's given query and returns the data from the results.
        """
        query_payload = json.dumps({
            'query': self._query,
            'variables': self._query_variables
        })
        headers = {
            'Content-Type': 'application/json',
            'X-Tableau-Auth': self._auth_token
        }
        params = {
            'headers': headers
        }
        if self._verify_request is not None:
            params['verify'] = self._verify_request

        response = requests.post(url=self._metadata_url, data=query_payload, **params)
        return response.json()['data']

    def execute(self) -> Iterator[Dict[str, Any]]:
        """
        Must be overriden by any extractor using this class. This should parse the result and yield each entity's
        metadata one by one.
        """
        pass

    def extract(self) -> Any:
        """
        Fetch one result at a time from the generator created by self.execute(), updating using the
        static record values if needed.
        """
        if not self._iterator:
            self._iterator = self.execute()

        try:
            record = next(self._iterator)
        except StopIteration:
            return None

        if self._static_dict:
            record.update(self._static_dict)

        return record


class TableauDashboardAuth:
    """
    Attempts to authenticate agains the Tableau REST API using the provided personal access token credentials.
    When successful, it will create a valid token that must be used on all subsequent requests.
    https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_concepts_auth.htm
    """

    API_BASE_URL = const.API_BASE_URL
    API_VERSION = const.API_VERSION
    SITE_NAME = const.SITE_NAME
    TABLEAU_ACCESS_TOKEN_NAME = const.TABLEAU_ACCESS_TOKEN_NAME
    TABLEAU_ACCESS_TOKEN_SECRET = const.TABLEAU_ACCESS_TOKEN_SECRET
    VERIFY_REQUEST = const.VERIFY_REQUEST

    def __init__(self, conf: ConfigTree) -> None:
        self._token: Optional[str] = None
        self._conf = conf
        self._access_token_name = self._conf.get_string(TableauDashboardAuth.TABLEAU_ACCESS_TOKEN_NAME)
        self._access_token_secret = self._conf.get_string(TableauDashboardAuth.TABLEAU_ACCESS_TOKEN_SECRET)
        self._api_version = self._conf.get_string(TableauDashboardAuth.API_VERSION)
        self._site_name = self._conf.get_string(TableauDashboardAuth.SITE_NAME, '')
        self._api_base_url = self._conf.get_string(TableauDashboardAuth.API_BASE_URL)
        self._verify_request = self._conf.get(TableauDashboardAuth.VERIFY_REQUEST, None)

    @property
    def token(self) -> Optional[str]:
        if not self._token:
            self._token = self._authenticate()
        return self._token

    def _authenticate(self) -> str:
        """
        Queries the auth/signin endpoint for the given Tableau instance using a personal access token.
        The API version differs with your version of Tableau.
        See https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_concepts_versions.htm
        for details or ask your Tableau server administrator.
        """
        self._auth_url = f"{self._api_base_url}/api/{self._api_version}/auth/signin"

        payload = json.dumps({
            'credentials': {
                'personalAccessTokenName': self._access_token_name,
                'personalAccessTokenSecret': self._access_token_secret,
                'site': {
                    'contentUrl': self._site_name
                }
            }
        })
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        # verify = False is needed bypass occasional (valid) self-signed cert errors. TODO: actually fix it!!
        params = {
            'headers': headers
        }
        if self._verify_request is not None:
            params['verify'] = self._verify_request

        response_json = requests.post(url=self._auth_url, data=payload, **params).json()
        return response_json['credentials']['token']



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import time
from os import listdir
from os.path import (
    basename, isfile, join, splitext,
)
from typing import (
    Dict, List, Optional, Type,
)

import pandas
from amundsen_rds.models import RDSModel
from amundsen_rds.models.base import Base
from pyhocon import ConfigFactory, ConfigTree
from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from databuilder.publisher.base_publisher import Publisher

LOGGER = logging.getLogger(__name__)


class MySQLCSVPublisher(Publisher):
    """
    A Publisher takes the table record folder as input and publishes csv to MySQL.
    The folder contains CSV file(s) for table records.

    The publish job works with rds models and SQLAlchemy ORM for data ingestion into MySQL.
    For more information:
    rds models: https://github.com/amundsen-io/amundsenrds
    SQLAlchemy ORM: https://docs.sqlalchemy.org/en/13/orm/
    """
    # Config keys
    # A directory that contains CSV files for records
    RECORD_FILES_DIR = 'record_files_directory'
    # It is used to provide unique tag to each record
    JOB_PUBLISH_TAG = 'job_publish_tag'

    # Connection string
    CONN_STRING = 'conn_string'
    # If its value is true, SQLAlchemy engine will log all statements
    ENGINE_ECHO = 'engine_echo'
    # Additional arguments used for engine
    CONNECT_ARGS = 'connect_args'
    # A transaction size that determines how often it commits.
    TRANSACTION_SIZE = 'transaction_size'
    # A progress report frequency that determines how often it report the progress.
    PROGRESS_REPORT_FREQUENCY = 'progress_report_frequency'

    _DEFAULT_CONFIG = ConfigFactory.from_dict({TRANSACTION_SIZE: 500,
                                               PROGRESS_REPORT_FREQUENCY: 500,
                                               ENGINE_ECHO: False})

    def __init__(self) -> None:
        super(MySQLCSVPublisher, self).__init__()

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(MySQLCSVPublisher._DEFAULT_CONFIG)

        self._count: int = 0
        self._progress_report_frequency = conf.get_int(MySQLCSVPublisher.PROGRESS_REPORT_FREQUENCY)
        self._record_files = self._list_files(conf, MySQLCSVPublisher.RECORD_FILES_DIR)
        self._sorted_record_files = self._sort_record_files(self._record_files)
        self._record_files_iter = iter(self._sorted_record_files)

        connect_args = {k: v for k, v in conf.get_config(MySQLCSVPublisher.CONNECT_ARGS,
                                                         default=ConfigTree()).items()}
        self._engine = create_engine(conf.get_string(MySQLCSVPublisher.CONN_STRING),
                                     echo=conf.get_bool(MySQLCSVPublisher.ENGINE_ECHO),
                                     connect_args=connect_args)
        self._session_factory = sessionmaker(bind=self._engine)
        self._transaction_size = conf.get_int(MySQLCSVPublisher.TRANSACTION_SIZE)

        self._publish_tag: str = conf.get_string(MySQLCSVPublisher.JOB_PUBLISH_TAG)
        if not self._publish_tag:
            raise Exception(f'{MySQLCSVPublisher.JOB_PUBLISH_TAG} should not be empty')

    def _list_files(self, conf: ConfigTree, path_key: str) -> List[str]:
        """
        List files from directory
        :param conf:
        :param path_key:
        :return: List of file paths
        """
        if path_key not in conf:
            return []

        path = conf.get_string(path_key)
        return [join(path, f) for f in listdir(path) if isfile(join(path, f))]

    def _sort_record_files(self, files: List[str]) -> List[str]:
        """
        Sort record files in the order of table dependencies
        :param files:
        :return:
        """
        sorted_table_names = [table.name for table in Base.metadata.sorted_tables]
        return sorted(files, key=lambda file: sorted_table_names.index(self._get_table_name_from_file(file)))

    def _get_table_name_from_file(self, file: str) -> str:
        """
        Get table name from file path
        :param file:
        :return:
        """
        try:
            filename = splitext(basename(file))[0]
            table_name, _ = filename.rsplit('_', 1)
            return table_name
        except Exception as e:
            LOGGER.exception(f'Error encountered while getting table name from file: {file}')
            raise e

    def publish_impl(self) -> None:
        """
        Publish records
        :return:
        """
        start = time.time()

        LOGGER.info(f'Publishing record files: {self._sorted_record_files}')
        session = self._session_factory()
        try:
            while True:
                try:
                    record_file = next(self._record_files_iter)
                    self._publish(record_file=record_file, session=session)
                except StopIteration:
                    break

            LOGGER.info(f'Committed total {self._count} statements')
            LOGGER.info(f'Successfully published. Elapsed: {time.time() - start} seconds')
        except Exception as e:
            LOGGER.exception('Failed to publish. Rolling back.')
            session.rollback()
            raise e
        finally:
            session.close()

    def _publish(self, record_file: str, session: Session) -> None:
        """
        Iterate over each row of the given csv file and convert each record to a rds model instance.
        Then the model instance will be inserted/updated in mysql.
        :param record_file:
        :param session:
        :return:
        """
        with open(record_file, 'r', encoding='utf8') as record_csv:
            table_name = self._get_table_name_from_file(record_file)
            table_model = self._get_model_from_table_name(table_name)
            if not table_model:
                raise RuntimeError(f'Failed to get model for table: {table_name}')

            for record_dict in pandas.read_csv(record_csv, na_filter=False).to_dict(orient='records'):
                record = self._create_record(model=table_model, record_dict=record_dict)
                session.merge(record)
                self._execute(session)
            session.commit()

    def _get_model_from_table_name(self, table_name: str) -> Optional[Type[RDSModel]]:
        """
        Get rds model for the given table name
        :param table_name:
        :return:
        """
        for model in Base._decl_class_registry.values():
            if hasattr(model, '__tablename__') and model.__tablename__ == table_name:
                return model
        return None

    def _create_record(self, model: Type[RDSModel], record_dict: Dict) -> RDSModel:
        """
        Convert the record dict to an instance of the given rds model
        and set additional attributes for the record instance
        :param model:
        :param record_dict:
        :return:
        """
        record = model(**record_dict)
        record.published_tag = self._publish_tag
        record.publisher_last_updated_epoch_ms = int(time.time() * 1000)
        return record

    def _execute(self, session: Session) -> None:
        """
        Commit pending record changes
        :param session:
        :return:
        """
        try:
            self._count += 1
            if self._count > 1 and self._count % self._transaction_size == 0:
                session.commit()
                LOGGER.info(f'Committed {self._count} records so far')

            if self._count > 1 and self._count % self._progress_report_frequency == 0:
                LOGGER.info(f'Processed {self._count} records so far')

        except Exception as e:
            LOGGER.exception('Failed to commit changes')
            raise e

    def get_scope(self) -> str:
        return 'publisher.mysql'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime


class FileMetadata(object):

    def __init__(self,
                 path: str,
                 last_updated: datetime,
                 size: int
                 ) -> None:
        self.path = path
        self.last_updated = last_updated
        self.size = size

    def __repr__(self) -> str:
        return f'FileMetadata(path={self.path!r}, last_updated={self.last_updated!r}, size={self.size!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import pkg_resources


def get_schema(schema: str) -> str:
    return pkg_resources.resource_string(__name__, schema).decode('utf-8')


application_schema = get_schema("schema/07_application_schema.json")
cluster_schema = get_schema("schema/00_cluster_schema.json")
schema_schema = get_schema("schema/01_schema_schema.json")
schema_cluster_relation = get_schema("schema/01_schema_cluster_relation.json")
database_schema = get_schema("schema/01_database_schema.json")
database_cluster_relation = get_schema("schema/01_database_cluster_relation.json")
table_schema = get_schema("schema/01_2_table_schema.json")
table_schema_relation = get_schema("schema/01_table_schema_relation.json")
source_schema = get_schema("schema/01_source_schema.json")
table_source_relation = get_schema("schema/01_table_source_relation.json")
bookmark_schema = get_schema("schema/01_3_bookmark.json")
report_schema = get_schema("schema/01_4_report.json")
column_schema = get_schema("schema/01_column_schema.json")
column_table_relation = get_schema("schema/01_column_table_relation.json")
lineage_schema = get_schema("schema/08_lineage_schema.json")
user_schema = get_schema("schema/02_user.json")
reader_schema = get_schema("schema/01_1_reader.json")
user_reader_relation = get_schema("schema/04_user_reader_relation.json")
reader_referenceable_relation = get_schema("schema/04_reader_referenceable_relation.json")
table_partition_schema = get_schema("schema/05_table_partition_schema.json")
hive_table_partition = get_schema("schema/05_1_hive_table_partition.json")
data_owner_schema = get_schema("schema/06_user_table_owner_relation.json")

# Dashboard definitions ------------------------------------------------------------------------------------------------

dashboard_group_schema = get_schema("schema/dashboard/01_group.json")
dashboard_schema = get_schema("schema/dashboard/02_dashboard.json")
dashboard_query_schema = get_schema("schema/dashboard/03_query.json")
dashboard_chart_schema = get_schema("schema/dashboard/04_chart.json")
dashboard_execution_schema = get_schema("schema/dashboard/05_execution.json")
dashboard_cluster_relation = get_schema("schema/dashboard/06_dashboard_cluster_relation.json")



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
import logging
from typing import List, Optional

LOGGER = logging.getLogger(__name__)


class Callback(object, metaclass=abc.ABCMeta):
    """
    A callback interface that expected to fire "on_success" if the operation is successful, else "on_failure" if
    operation failed.
    """

    @abc.abstractmethod
    def on_success(self) -> None:
        """
        A call back method that will be called when operation is successful
        :return: None
        """
        pass

    @abc.abstractmethod
    def on_failure(self) -> None:
        """
        A call back method that will be called when operation failed
        :return: None
        """
        pass


def notify_callbacks(callbacks: List[Callback], is_success: bool) -> None:
    """
    A Utility method that notifies callback. If any callback fails it will still go through all the callbacks,
    and raise the last exception it experienced.

    :param callbacks:
    :param is_success:
    :return:
    """

    if not callbacks:
        LOGGER.info('No callbacks to notify')
        return

    LOGGER.info('Notifying callbacks')

    last_exception: Optional[Exception] = None
    for callback in callbacks:
        try:
            if is_success:
                callback.on_success()
            else:
                callback.on_failure()
        except Exception as e:
            LOGGER.exception('Failed while notifying callback')
            last_exception = e

    if last_exception:
        raise last_exception



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, List, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasCommonTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.badge import Badge as RDSBadge

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


class Badge:
    def __init__(self, name: str, category: str):
        # Amundsen UI always formats badge display with first letter capitalized while other letters are lowercase.
        # Clicking table badges in UI always results in searching lower cases badges
        # https://github.com/amundsen-io/amundsen/blob/6ec9b398634264e52089bb9e1b7d76a6fb6a35a4/frontend/amundsen_application/static/js/components/BadgeList/index.tsx#L56
        # If badges stored in neo4j are not lowercase, they won't be searchable in UI.
        self.name = name.lower()
        self.category = category.lower()

    def __repr__(self) -> str:
        return f'Badge({self.name!r}, {self.category!r})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Badge):
            return NotImplemented
        return self.name == other.name and self.category == other.category


class BadgeMetadata(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    Badge model.
    """
    BADGE_NODE_LABEL = 'Badge'
    BADGE_KEY_FORMAT = '{badge}'
    BADGE_CATEGORY = 'category'

    # Relation between entity and badge
    BADGE_RELATION_TYPE = 'HAS_BADGE'
    INVERSE_BADGE_RELATION_TYPE = 'BADGE_FOR'

    LABELS_PERMITTED_TO_HAVE_BADGE = ['Table', 'Dashboard', 'Column', 'Feature', 'Type_Metadata']

    def __init__(self,
                 start_label: str,
                 start_key: str,
                 badges: List[Badge],
                 ):
        if start_label not in BadgeMetadata.LABELS_PERMITTED_TO_HAVE_BADGE:
            raise Exception(f'badges for {start_label} are not supported')
        self.start_label = start_label
        self.start_key = start_key
        self.badges = badges

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def __repr__(self) -> str:
        return f'BadgeMetadata({self.start_label!r}, {self.start_key!r}, {self.badges!r})'

    def create_next_node(self) -> Optional[GraphNode]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    @staticmethod
    def get_badge_key(name: str) -> str:
        if not name:
            return ''
        return BadgeMetadata.BADGE_KEY_FORMAT.format(badge=name)

    def get_badge_nodes(self) -> List[GraphNode]:
        nodes = []
        for badge in self.badges:
            if badge:
                node = GraphNode(
                    key=self.get_badge_key(badge.name),
                    label=self.BADGE_NODE_LABEL,
                    attributes={
                        self.BADGE_CATEGORY: badge.category
                    }
                )
                nodes.append(node)
        return nodes

    def get_badge_relations(self) -> List[GraphRelationship]:
        relations = []
        for badge in self.badges:
            relation = GraphRelationship(
                start_label=self.start_label,
                end_label=self.BADGE_NODE_LABEL,
                start_key=self.start_key,
                end_key=self.get_badge_key(badge.name),
                type=self.BADGE_RELATION_TYPE,
                reverse_type=self.INVERSE_BADGE_RELATION_TYPE,
                attributes={}
            )
            relations.append(relation)
        return relations

    def get_badge_records(self) -> List[RDSModel]:
        records = []
        for badge in self.badges:
            if badge:
                record = RDSBadge(
                    rk=self.get_badge_key(badge.name),
                    category=badge.category
                )
                records.append(record)

        return records

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create badge nodes
        :return:
        """
        nodes = self.get_badge_nodes()
        for node in nodes:
            yield node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relations = self.get_badge_relations()
        for relation in relations:
            yield relation

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        records = self.get_badge_records()
        for record in records:
            yield record

    def _create_atlas_classification_entity(self, badge: Badge) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, badge.name),
            ('category', badge.category),
            ('name', badge.name)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.badge,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def _create_atlas_classification_relation(self, badge: Badge) -> AtlasRelationship:
        table_relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.badge,
            entityType1=AtlasCommonTypes.data_set,
            entityQualifiedName1=self.start_key,
            entityType2=AtlasRelationshipTypes.badge,
            entityQualifiedName2=badge.name,
            attributes={}
        )

        return table_relationship

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        for badge in self.badges:
            yield self._create_atlas_classification_relation(badge)

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        for badge in self.badges:
            yield self._create_atlas_classification_entity(badge)

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, Dict

from pyhocon import ConfigTree

from databuilder.transformer.base_transformer import Transformer

CALLBACK_FUNCTION = 'callback_function'
FIELD_NAME = 'field_name'

LOGGER = logging.getLogger(__name__)


class GenericTransformer(Transformer):
    """
    A generic transformer that accepts a callback function that transforms the record on specified field.
    """

    def init(self, conf: ConfigTree) -> None:
        self._callback_function = conf.get(CALLBACK_FUNCTION)
        self._field_name = conf.get_string(FIELD_NAME)

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:

        for k, v in record.items():
            if k == self._field_name:
                new_val = self._callback_function(v)
                record[k] = new_val
        return record

    def get_scope(self) -> str:
        return 'transformer.generic'
